
from epyk.core.html import Html
from epyk.core.html import Defaults
from epyk.core.html.options import OptPanel
from epyk.core.html.options import OptText
#
from epyk.core.js import JsUtils
from epyk.core.js.html import JsHtml
from epyk.core.js.html import JsHtmlPanels

# The list of CSS classes#
from epyk.core.css import Defaults as css_defaults
from epyk.core.css.styles import GrpClsContainer


class Panel(Html.Html):
  name, category, callFnc = 'Panel', 'Layouts', 'panel'

  def __init__(self, report, htmlObj, title, color, width, height, htmlCode, helper, options, profile):
    if isinstance(htmlObj, list) and htmlObj:
      for obj in htmlObj:
        if hasattr(obj, 'inReport'):
          obj.inReport = False
    elif htmlObj is not None and hasattr(htmlObj, 'inReport'):
      htmlObj.inReport = False # Has to be defined here otherwise it is set to late
    component = []
    if title is not None:
      self.title = report.ui.title(title)
      self.title.inReport = False
      component.append(self.title)
    container = report.ui.div(htmlObj)
    container.inReport = False
    component.append(container)
    self.add_helper(helper)
    super(Panel, self).__init__(report, component, code=htmlCode, profile=profile,
                                css_attrs={"color": color, "width": width, "height": height})
    container.set_attrs(name="name", value="panel_%s" % self.htmlId)

  @property
  def style(self):
    """
    Description:
    ------------

    :rtype: GrpClsContainer.ClassDiv
    """
    if self._styleObj is None:
      self._styleObj = GrpClsContainer.ClassDiv(self)
    return self._styleObj

  def __add__(self, htmlObj):
    """ Add items to a container """
    htmlObj.inReport = False # Has to be defined here otherwise it is set to late
    self.val.append(htmlObj)
    return self

  @property
  def dom(self):
    """
    Description:
    ------------
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    :return: A Javascript Dom object

    :rtype: JsHtmlPanels.JsHtmlPanel
    """
    if self._dom is None:
      self._dom = JsHtmlPanels.JsHtmlPanel(self, report=self._report)
    return self._dom

  def __str__(self):
    str_div = "".join([v.html() if hasattr(v, 'html') else str(v) for v in self.val])
    return "<div %s>%s</div>%s" % (self.get_attrs(pyClassNames=self.style.get_classes()), str_div, self.helper)


class PanelSplit(Html.Html):
  __reqJs, __reqCss = ['jqueryui'], ['jqueryui']
  name, category, callFnc = 'Panel Split', 'Layouts', 'panelsplit'

  def __init__(self, report, width, height, left_width, left_obj, right_obj, resizable, helper, profile):
    super(PanelSplit, self).__init__(report, None, css_attrs={"width": width, "height": height, 'white-space': 'nowrap'}, profile=profile)
    self.left_width, self.resizable = left_width, resizable
    if left_obj is not None:
      self.left(left_obj)
    if right_obj is not None:
      self.right(right_obj)
    self.css_left = {'flex': '0 0 auto', 'overflow': 'auto', 'padding': '5px', 'min-width': '100px', 'width': "%s%s" % (self.left_width[0], self.left_width[1]),
                     'white-space': 'nowrap'}
    self.css_right = {'flex': '0 1 auto', 'overflow': 'auto', 'padding': '5px', 'width': '100%', 'background': self._report.theme.greys[0],
                      'border-left': '3px solid %s' % self._report.theme.success[1]}
    self.css({'display': 'flex', 'flex-direction': 'row', 'overflow': 'hidden', 'xtouch-action': 'none'})

  def left(self, html_obj):
    """
    Description:
    ------------
    Add the left component to the panel

    Attributes:
    ----------
    :param html_obj:
    """
    html_obj.inReport = False
    self.html_left = html_obj
    return self

  def right(self, html_obj):
    """
    Description:
    ------------
    Add the right component to the panel

    Attributes:
    ----------
    :param html_obj:
    """
    html_obj.inReport = False
    self.html_right = html_obj
    return self

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).extend([
      '$("#%(htmlId)s_left").resizable({handleSelector: ".splitter", resizeHeight: false});' % {'htmlId': self.htmlId},
      '$("#%(htmlId)s_right").resizable({handleSelector: ".splitter-horizontal", resizeWidth: true})' % {
        'htmlId': self.htmlId}])
    return '''
      <div %(attrs)s>
        <div style="%(css_left)s" id="%(htmlId)s_left" class="panel-left">%(left)s</div>
        <div style="%(css_right)s" id="%(htmlId)s_right" class="panel-right">%(right)s</div>
      </div>
      ''' % {"attrs": self.get_attrs(pyClassNames=self.style.get_classes()), "htmlId": self.htmlId, 'left': self.html_left.html(),
             'right': self.html_right.html(), 'css_left': css_defaults.inline(self.css_left), 'css_right': css_defaults.inline(self.css_right)}


class PanelSlide(Panel):
  __reqCss, __reqJs = ['font-awesome'], ['font-awesome']
  name, category, callFnc = 'Slide Panel', 'Panels', 'slide'

  def __init__(self, report, htmlObj, title, color, width, height, htmlCode, helper, options, profile):
    super(PanelSlide, self).__init__(report, htmlObj, None, color, width, height, htmlCode, helper, options, profile)
    self.add_helper(helper)
    self.icon = self._report.ui.icon("").css({"display": 'inline-block', 'margin': '0 2px',
                                              'line-height': "%spx" % Defaults.LINE_HEIGHT, 'font-size': "%spx" % Defaults.BIG_ICONS})
    self.text = self._report.ui.title(title).css({"display": 'inline-block', 'margin': 0})
    self.title = self._report.ui.div([self.icon, self.text])
    self.title.inReport = False
    self.title.css({"cursor": 'pointer', "padding": "0 2px 0 0"})
    self._vals, self.__clicks = [self.title] + self._vals, []
    self.__options = OptPanel.OptionPanelSliding(report, options)

  @property
  def options(self):
    """
    Description:
    ------------

    :rtype: OptPanel.OptionPanelSliding
    """
    return self.__options

  def click(self, jsFncs, profile=False):
    """
    Description:
    ------------
    Event added to the title bar.
    This will be triggered first

    Attributes:
    ----------
    :param jsFncs:
    :param profile:
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self.__clicks = jsFncs
    return self

  def __add__(self, htmlObj):
    """ Add items to a container """
    self.val[1] += htmlObj
    return self

  def __str__(self):
    if self.options.expanded:
      icon_change = self.options.icon_closed
      icon_current = self.options.icon_expanded
      self.icon.set_icon(self.options.icon_expanded)
    else:
      icon_change = self.options.icon_expanded
      icon_current = self.options.icon_closed
      self._vals[1].style.css.display = 'none'
      self.icon.set_icon(self.options.icon_closed)
    self.title.click(self.__clicks + [
      self._report.js.getElementsByName("panel_%s" % self.htmlId).first.toggle(),
      self.icon.dom.switchClass(icon_current, icon_change)])
    str_div = "".join([v.html() if hasattr(v, 'html') else str(v) for v in self.val])
    return "<div %s>%s</div>%s" % (self.get_attrs(pyClassNames=self.style.get_classes()), str_div, self.helper)


class Div(Html.Html):
  name, category, callFnc = 'Simple Container', 'Layouts', 'div'

  def __init__(self, report, htmlObj, label, color, width, icon, height, editable, align, padding, htmlCode, tag,
               helper, options, profile):
    if isinstance(htmlObj, list) and htmlObj:
      newHtmlObj = []
      for obj in htmlObj:
        if isinstance(obj, list) and obj:
          newHtmlObj.append(report.ui.div(obj, label, color, width, icon, height, editable, align, padding, htmlCode, tag, helper, profile))
        else:
          newHtmlObj.append(obj)
        if hasattr(newHtmlObj[-1], 'inReport'):
          if options.get('inline'):
            newHtmlObj[-1].style.css.display = 'inline-block'
          newHtmlObj[-1].inReport = False
          #if newHtmlObj[-1].css("display") != 'None':
          #  newHtmlObj[-1].css({"display": 'inline-block'})
      htmlObj = newHtmlObj
    elif htmlObj is not None and hasattr(htmlObj, 'inReport'):
      htmlObj.inReport = False # Has to be defined here otherwise it is set to late
    super(Div, self).__init__(report, htmlObj, code=htmlCode, css_attrs={"color": color, "width": width, "height": height},
                              profile=profile)
    self.tag = tag
    self.__options = OptPanel.OptionsDiv(self, options)
    # Add the component predefined elements
    self.add_icon(icon)
    self.add_label(label)
    self.add_helper(helper)

    self.css({'text-align': align})
    if padding is not None:
      self.css('padding', '%s' % padding)
    if editable:
      self.set_attrs(name='contenteditable', value="true")
      self.css('overflow', 'auto')

  @property
  def dom(self):
    """
    Description:
    ------------
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    :return: A Javascript Dom object

    :rtype: JsHtml.JsHtmlRich
    """
    if self._dom is None:
      self._dom = JsHtml.JsHtmlRich(self, report=self._report)
    return self._dom

  def __add__(self, htmlObj):
    """ Add items to a container """
    htmlObj.inReport = False # Has to be defined here otherwise it is set to late
    if self.options.inline:
      htmlObj.style.css.display = 'inline-block'
    if not isinstance(self.val, list):
      self._vals = [self.val]
    self.val.append(htmlObj)
    return self

  @property
  def options(self):
    """
    Description:
    ------------
    Property to set all the possible object for a button

    :rtype: OptPanel.OptionsDiv
    """
    return self.__options

  def __getitem__(self, i):
    return self.val[i]

  @property
  def style(self):
    """

    :rtype: GrpClsContainer.ClassDiv
    """
    if self._styleObj is None:
      self._styleObj = GrpClsContainer.ClassDiv(self)
    return self._styleObj

  def build(self, data=None, options=None, profile=False):
    if isinstance(data, dict):
      # check if there is no nested HTML components in the data
      js_data = "{%s}" % ",".join(["%s: %s" % (k, JsUtils.jsConvertData(v, None)) for k, v in data.items()])
    else:
      js_data = JsUtils.jsConvertData(data, None)
    options, js_options = options or {}, []
    for k, v in options.items():
      if isinstance(v, dict):
        row = ["%s: %s" % (s_k, JsUtils.jsConvertData(s_v, None)) for s_k, s_v in v.items()]
        js_options.append("%s: {%s}" % (k, ", ".join(row)))
      else:
        js_options.append("%s: %s" % (k, JsUtils.jsConvertData(v, None)))
    return "%s.innerHTML = %s" % (self.dom.varId, js_data) #, "{%s}" % ",".join(js_options))

  def __str__(self):
    str_div = "".join([v.html() if hasattr(v, 'html') else str(v) for v in self.val])
    return "<div %s>%s</div>%s" % (self.get_attrs(pyClassNames=self.style.get_classes()), str_div, self.helper)

  # -----------------------------------------------------------------------------------------
  #                                    EXPORT OPTIONS
  # -----------------------------------------------------------------------------------------
  def to_word(self, document):
    if isinstance(self.vals, list):
      for val in self.vals:
        if hasattr(val, 'inReport'):
          val.to_word(document)
    else:
      if hasattr(self.vals, 'inReport'):
        self.vals.to_word(document)


class Td(Html.Html):
  name, category, callFnc = 'Column', 'Layouts', 'col'

  def __init__(self, report, htmlObjs, header, position, width, height, align, options, profile):
    self.position, self.rows_css, self.row_css_dflt, self.header = position, {}, {}, header
    super(Td, self).__init__(report, [], css_attrs={"width": width, "height": height, 'white-space': 'nowrap'}, profile=profile)
    self.__options = options
    self.attr["align"] = options.cell_align or align
    if htmlObjs is not None:
      for htmlObj in htmlObjs:
        self.__add__(htmlObj)

  def __add__(self, htmlObj):
    """ Add items to a container """
    if hasattr(htmlObj, 'inReport'):
      htmlObj.inReport = False # Has to be defined here otherwise it is set to late
    self.val.append(htmlObj)
    return self

  def colspan(self, i):
    """
    Description:
    ------------
    The colspan attribute defines the number of columns a cell should span.

    Related Pages:

			https://www.w3schools.com/tags/att_td_colspan.asp

    Attributes:
    ----------
    :param i:
    """
    self.attr['colspan'] = i
    return self

  def rowspan(self, i):
    """
    Description:
    ------------
    The rowspan attribute specifies the number of rows a cell should span.

    Related Pages:

			https://www.w3schools.com/tags/att_td_rowspan.asp

    Attributes:
    ----------
    :param i:
    """
    self.attr['rowspan'] = i
    return self

  def __getitem__(self, i):
    return self.val[i]

  def __str__(self):
    content = [htmlObj.html() if hasattr(htmlObj, 'inReport') else str(htmlObj) for htmlObj in self.val]
    if self.header:
      return '<th %s>%s</th>' % (self.get_attrs(pyClassNames=self.style.get_classes()), "".join(content))

    return '<td %s>%s</td>' % (self.get_attrs(pyClassNames=self.style.get_classes()), "".join(content))


class Tr(Html.Html):
  name, category, callFnc = 'Column', 'Layouts', 'col'

  def __init__(self, report, htmlObjs, header, position, width, height, align, options, profile):
    self.position, self.header = position, header
    super(Tr, self).__init__(report, [], css_attrs={"width": width, "height": height, 'text-align': align}, profile=profile)
    self.__options = options
    if htmlObjs is not None:
      for htmlObj in htmlObjs:
        self.__add__(htmlObj)
    self.style.justify_content = self.position

  def __add__(self, htmlObj):
    """ Add items to a container """
    if not isinstance(htmlObj, Td):
      if not isinstance(htmlObj, list):
        htmlObj = [htmlObj]
      htmlObj = Td(self._report, htmlObj, self.header, None, (None, "%"), (None, "%"), 'center', self.__options, False)
    self.val.append(htmlObj)
    return self

  @property
  def dom(self):
    """
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    :return: A Javascript Dom object

    :rtype: JsHtmlPanels.JsHtmlTr
    """
    if self._dom is None:
      self._dom = JsHtmlPanels.JsHtmlTr(self, report=self._report)
    return self._dom

  def __getitem__(self, i):
    """
    Description:
    ------------
    Return the internal column in the row for the given index

    Attributes:
    ----------
    :param i: the column index

    :rtype: Td
    """
    return self.val[i]

  def __str__(self):
    cols = [htmlObj.html() for i, htmlObj in enumerate(self.val)]
    return '<tr %s>%s</tr>' % (self.get_attrs(pyClassNames=self.style.get_classes()), "".join(cols))


class Caption(Html.Html):

  def __init__(self, report, text, color, align, width, height, htmlCode, tooltip, options, profile):
    super(Caption, self).__init__(report, text, css_attrs={"width": width, "height": height, "color": color, 'text-align': align},
                               code=htmlCode, profile=profile)
    self.__options = OptText.OptionsText(self, options)
    if tooltip is not None:
      self.tooltip(tooltip)

  @property
  def options(self):
    """
    Description:
    ------------
    Property to set all the possible object for a button

    :rtype: OptText.OptionsText
    """
    return self.__options

  def __str__(self):
    val = self._report.py.markdown.all(self.val) if self.options.showdown else self.val
    return '<caption %s>%s</caption>' % (self.get_attrs(pyClassNames=self.style.get_classes()), self.val)


class TSection(Html.Html):

  def __init__(self, report, type, rows=None, options=None, profile=None):
    super(TSection, self).__init__(report, [])
    self.__section = type
    self.__options = OptPanel.OptionPanelTable(report, options)
    if rows is not None:
      for row in rows:
        self.__add__(row)

  @property
  def options(self):
    """
    Description:
    ------------

    :rtype: OptPanel.OptionPanelTable
    """
    return self.__options

  def __getitem__(self, i):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param i: Integer. The internal row based on the index

    :rtype: Tr
    """
    return self.val[i]

  def __add__(self, row_data):
    """ Add items to a container """
    if not isinstance(row_data, Tr):
      row_data = Tr(self._report, row_data, self.__section == 'thead', None, (100, "%"), (100, "%"), 'center', self.options, False)

    self.val.append(row_data)
    return self

  def __str__(self):
    cols = "".join([htmlObj.html() for i, htmlObj in enumerate(self.val)])
    return '<%(section)s %(attr)s>%(cols)s</%(section)s>' % {'section': self.__section, 'cols': cols,
                  'attr': self.get_attrs(pyClassNames=self.style.get_classes())}


class Table(Html.Html):
  name, category, callFnc = 'Table', 'Layouts', 'table'

  def __init__(self, report, rows, width, height, helper, options, profile):
    super(Table, self).__init__(report, [], css_attrs={"width": width, "height": height, 'table-layout': 'auto',
            'white-space': 'nowrap', 'border-collapse': 'collapse', 'box-sizing': 'border-box'}, profile=profile)
    self.add_helper(helper, css={"float": "none", "margin-left": "5px"})
    self.__options = OptPanel.OptionPanelTable(report, options)
    self.header, self.body, self.footer = TSection(self._report, 'thead', options=options), TSection(self._report, 'tbody', options=options), TSection(self._report, 'tfoot', options=options)
    self.caption = None
    if rows is not None:
      for row in rows:
        self.__add__(row)

  @property
  def options(self):
    """
    Description:
    ------------

    :rtype: OptPanel.OptionPanelTable
    """
    return self.__options

  def __add__(self, row_data):
    """ Add items to a container """
    if isinstance(row_data, Tr):
      row = row_data
    else:
      if not self.header.val and not self.body.val and self.options.header:
        row = Tr(self._report, row_data, True, None, (100, "%"), (100, "%"), 'center', self.options, False)
      else:
        row = Tr(self._report, row_data, False, None, (100, "%"), (100, "%"), 'left', self.options, False)
    if row.header:
      self.header += row
    else:
      self.body += row
    return self

  def from_array(self, data, dim):
    """
    Description:
    ------------
    Load data from a 2D array

    Attributes:
    ----------
    :param data: Array. The list of data
    :param dim: Integer. The number of columns in the table
    """
    v_count = len(data)
    modulo_rest = v_count % dim
    modulo_result = v_count // dim
    for i in range(0, modulo_result):
      row = [data[i * dim + j] for j in range(0, dim)]
      self += row
    if modulo_rest:
      self += data[-modulo_rest:]

  def line(self, text="&nbsp;", align="left", dim=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param text:
    :param align:
    :param dim: Integer. The number of columns in the table
    """
    cell = Td(self._report, [text], False, None, (None, "%"), (None, "%"), align, self.options, False)
    cell.colspan(dim or len(self.body.val[0].val))
    self += Tr(self._report, [cell], False, None, (100, "%"), (100, "%"), align, self.options, False)
    return cell

  def add_caption(self, text, color=None, align=None, width=(100, "%"), height=(100, "%"), htmlCode=None, tooltip=None,
                  options=None, profile=False):
    """
    Description:
    ------------
    The <caption> tag defines a table caption.

    Related Pages:

			https://www.w3schools.com/tags/tag_caption.asp

    Attributes:
    ----------
    :param text:
    :param color:
    :param align:
    :param width:
    :param height:
    :param htmlCode:
    :param tooltip:
    :param options:
    :param profile:
    """
    self.caption = Caption(self._report, text, color, align, width, height, htmlCode, tooltip, options, profile)
    return self.caption

  def get_header(self, i=0):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param i: Integer.

    :return:
    """
    return self.header.val[i]

  def get_footer(self, i=0):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param i: Integer.

    :return:
    """
    return self.footer.val[i]

  def col(self, i):
    """

    :param i:
    """
    cells = []
    if self.header:
      for h in self.header:
        cells.append(h[i])
    if self.body:
      for b in self.body:
        cells.append(b[i])
    if self.footer:
      for f in self.footer:
        cells.append(f[i])
    for c in cells:
      yield c

  def __getitem__(self, i):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param i: Integer. The internal row based on the index

    :rtype: Tr
    """
    if not self.body.val:
      return []

    return self.body.val[i]

  def __str__(self):
    caption = "" if self.caption is None else self.caption.html()
    return '<table %s>%s%s%s%s</table>%s' % (self.get_attrs(pyClassNames=self.style.get_classes()), caption, self.header.html(), self.body.html(), self.footer.html(), self.helper)


class Col(Html.Html):
  name, category, callFnc = 'Column', 'Layouts', 'col'

  def __init__(self, report, htmlObjs, position, width, height, align, helper, profile):
    self.position,  self.rows_css, self.row_css_dflt = position, {}, {}
    super(Col, self).__init__(report, [], css_attrs={"width": width, "height": height}, profile=profile)
    if htmlObjs is not None:
      for htmlObj in htmlObjs:
        self.__add__(htmlObj)
    if align == "center":
      self.css({'margin': 'auto', 'display': 'inline-block', 'text-align': 'center'})
    else:
      self.css({'display': 'inline-block'})
    self.attr["class"].add('col')
    self.style.justify_content = self.position

  def __add__(self, htmlObj):
    """
    Description:
    ------------
    Add items to a container

    Attributes:
    ----------
    :param htmlObj:

    :return:
    """
    if not hasattr(htmlObj, 'inReport'):
      htmlObj = self._report.ui.div(htmlObj)
    htmlObj.inReport = False # Has to be defined here otherwise it is set to late
    self.val.append(htmlObj)
    return self

  def build(self, data=None, options=None, profile=False):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param data:
    :param options:
    :param profile:
    """
    return self.val[0].build(data, options, profile)

  def set_size(self, n):
    """
    Description:
    ------------
    Set the column size

    Usage::

      ps = rptObj.ui.layouts.grid()
    ps += [rptObj.ui.text("test %s" % i) for i in range(5)]
    ps[0][0].set_size(10)

    Attributes:
    ----------

    """
    self.attr["class"].add("col-%s" % n)
    return self

  def __getitem__(self, i):
    return self.val[i]

  def __str__(self):
    content = [htmlObj.html() for htmlObj in self.val]
    return '<div %s>%s</div>' % (self.get_attrs(pyClassNames=self.style.get_classes()), "".join(content))

  # -----------------------------------------------------------------------------------------
  #                                    EXPORT OPTIONS
  # -----------------------------------------------------------------------------------------
  def to_word(self, document):
    for i, htmlObj in enumerate(self.vals):
      htmlObj.to_word(document)

  def to_xls(self, workbook, worksheet, cursor):
    for i, htmlObj in enumerate(self.vals):
      try:
        htmlObj.to_xls(workbook, worksheet, cursor)
      except Exception as err:
        cell_format = workbook.add_format({'bold': True, 'font_color': 'red'})
        worksheet.write(cursor['row'], 0, str(err), cell_format)
        cursor['row'] += 2


class Row(Html.Html):
  name, category, callFnc = 'Column', 'Layouts', 'col'
  __reqCss, __reqJs = ['bootstrap'], ['bootstrap']

  def __init__(self, report, htmlObjs, position, width, height, align, helper, profile):
    self.position = position
    super(Row, self).__init__(report, [], css_attrs={"width": width, "height": height}, profile=profile)
    if htmlObjs is not None:
      for htmlObj in htmlObjs:
        self.__add__(htmlObj)
    self.attr["class"].add('row')
    self.attr["class"].add('no-gutters')
    self.style.justify_content = self.position

  @property
  def dom(self):
    """
    Description:
    ------------
    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    :return: A Javascript Dom object

    :rtype: JsHtmlPanels.JsHtmlRow
    """
    if self._dom is None:
      self._dom = JsHtmlPanels.JsHtmlRow(self, report=self._report)
    return self._dom

  def __add__(self, htmlObj):
    """ Add items to a container """
    if not isinstance(htmlObj, Col):
      if not isinstance(htmlObj, list):
        htmlObj = [htmlObj]
      htmlObj = self._report.ui.layouts.col(htmlObj)
    htmlObj.inReport = False # Has to be defined here otherwise it is set to late
    self.val.append(htmlObj)
    return self

  def __getitem__(self, i):
    """
    Return the internal column in the row for the given index

    :param i: the column index
    :rtype: Col
    """
    return self.val[i]

  def __str__(self):
    cols = []
    for i, htmlObj in enumerate(self.val):
      cols.append(htmlObj.html())
    return '<div %s>%s</div>' % (self.get_attrs(pyClassNames=self.style.get_classes()), "".join(cols))


class Grid(Html.Html):
  name, category, callFnc = 'Grid', 'Layouts', 'grid'
  __reqCss, __reqJs = ['bootstrap'], ['bootstrap']

  def __init__(self, report, rows, width, height, align, helper, profile):
    super(Grid, self).__init__(report, [], css_attrs={"width": width, "height": height}, profile=profile)
    self.css({'overflow-x': 'hidden', 'padding': 0})
    self.attr["class"].add("container-fluid")
    if align == 'center':
      self.css({'margin': 'auto'})
    if rows is not None:
      for row in rows:
        self.__add__(row)

  def __add__(self, row_data):
    """ Add items to a container """
    if isinstance(row_data, Row):
      row = row_data
    else:
      row = self._report.ui.layouts.row()
      for htmlObjWithDim in row_data:
        if isinstance(htmlObjWithDim, tuple):
          htmlObj, dim = htmlObjWithDim
        else:
          htmlObj, dim = htmlObjWithDim, None
        row += htmlObj
        if dim is not None:
          row[-1].attr["class"].add("col-%s" % dim)
    row.inReport = False
    self.val.append(row)
    #self.colsAlign.append("left")
    return self

  def __getitem__(self, i):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param i: Integer. The internal row based on the index

    :rtype: Row
    """
    return self.val[i]

  @property
  def dom(self):
    """
    Description:
    ------------
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    :return: A Javascript Dom object

    :rtype: JsHtmlPanels.JsHtmlGrid
    """
    if self._dom is None:
      self._dom = JsHtmlPanels.JsHtmlGrid(self, report=self._report)
    return self._dom

  def resize(self):
    """
    Description:
    ------------
    For the resizing of the space for the containers.

    This will rescale based on the number of items and the fact that the max per row is 12
    It will update the colsDim list
    """
    max_size = int(12 / len(self.colsDim))
    self.colsDim = [max_size] * len(self.colsDim)
    return self

  def __str__(self):
    rows = [htmlObj.html() for htmlObj in self.val]
    return '<div %s>%s</div>' % (self.get_attrs(pyClassNames=self.style.get_classes()), "".join(rows))


  # -----------------------------------------------------------------------------------------
  #                                    EXPORT OPTIONS
  # -----------------------------------------------------------------------------------------
  def to_word(self, document):
    for i, htmlObj in enumerate(self.vals):
      try:
        htmlObj.to_word(document)
      except Exception as err:
        from docx.shared import RGBColor

        errotTitle = document.add_heading().add_run("Error")
        errotTitle.font.color.rgb = RGBColor(255, 0, 0)
        errotTitle.font.italic = True
        errorParagraph = document.add_paragraph().add_run((str(err)))
        errorParagraph.font.color.rgb = RGBColor(255, 0, 0)
        errorParagraph.font.italic = True

  def to_xls(self, workbook, worksheet, cursor):
    for i, htmlObj in enumerate(self.vals):
      try:
        htmlObj.to_xls(workbook, worksheet, cursor)
      except Exception as err:
        cell_format = workbook.add_format({'bold': True, 'font_color': 'red'})
        worksheet.write(cursor['row'], 0, str(err), cell_format)
        cursor['row'] += 2

  # -----------------------------------------------------------------------------------------
  #                                    MARKDOWN SECTION
  # -----------------------------------------------------------------------------------------
  @staticmethod
  def matchMarkDown(val): return True if val == "[" else None

  @classmethod
  def convertMarkDown(cls, val, regExpResult, report=None):
    pass

  @classmethod
  def jsMarkDown(self, vals):
    return '''
      '''


class Tabs(Html.Html):
  name, category, callFnc = 'Tabs', 'Layouts', 'tabs'

  def __init__(self, report, color, width, height, htmlCode, helper, options, profile):
    super(Tabs, self).__init__(report, "", code=htmlCode, css_attrs={"width": width, "height": height, 'color': color}, profile=profile)
    self.__panels, self.__panel_objs, self.__selected = [], {}, None
    self.tabs_name, self.panels_name = "button_%s" % self.htmlId, "panel_%s" % self.htmlId
    self.tabs_container = self._report.ui.div([])
    self.tabs_container.inReport = False
    self.add_helper(helper)
    self.__options = OptPanel.OptionPanelTabs(report, options)

  @property
  def options(self):
    """
    Description:
    ------------

    :rtype: OptPanel.OptionPanelTabs
    """
    return self.__options

  @property
  def dom(self):
    """
    Description:
    ------------
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    :return: A Javascript Dom object

    :rtype: JsHtmlPanels.JsHtmlTabs
    """
    if self._dom is None:
      self._dom = JsHtmlPanels.JsHtmlTabs(self, report=self._report)
    return self._dom

  def __getitem__(self, name):
    return self.__panel_objs[name]

  def select(self, name):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param name:
    """
    self.__selected = name
    return self

  def panel(self, name):
    """
    Description:
    ------------
    Get the panel obkect

    Attributes:
    ----------
    :param name: String. The tab name

    :rtype: Div
    """
    return self.__panel_objs[name]["content"]

  def tab(self, name):
    """
    Description:
    ------------
    Get the tab container

    Attributes:
    ----------
    :param name: String. The tab name

    :rtype: Div
    """
    return self.__panel_objs[name]["tab"][0]

  def tab_holder(self, name):
    """
    Description:
    ------------
    Get the tab container

    Attributes:
    ----------
    :param name: String. The tab name

    :rtype: Div
    """
    return self.__panel_objs[name]["tab"]

  def tabs(self):
    """
    Description:
    ------------
    Get the tab container

    Attributes:
    ----------
    :param name: String. The tab name

    :rtype: Div
    """
    for tab_obj in self.__panel_objs.values():
      yield tab_obj["tab"]

  def add_panel(self, name, div, icon=None, selected=False, css_tab=None, css_tab_clicked=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param name:
    :param div:
    :param selected: Boolean. Flag to set the selected panel
    :param css_tab:
    """
    div.css({"display": 'none'})
    div.inReport = False
    div.set_attrs(name="name", value=self.panels_name)
    self.__panels.append(name)
    if icon is not None:
      tab = self._report.ui.div([
        self._report.ui.icon(icon).css({"display": 'block', "width": '100%', "font-size": css_defaults.font(4)}),
        name], width=("100", "px"))
    else:
      tab = self._report.ui.div(name, width=("100", "px"))
    tab_style = self.options.tab_style(name, css_tab)
    tab_style_clicked = self.options.tab_clicked_style(name, css_tab_clicked)
    tab.css(tab_style).css({"padding": '5px 0'})
    tab.set_attrs(name="name", value=self.tabs_name)
    tab.set_attrs(name="data-index", value=len(self.__panels) - 1)
    tab_container = self._report.ui.div(tab, width=("100", "px"))
    tab_container.inReport = False
    tab_container.css({'display': 'inline-block'})
    css_cls_name = None
    tab.click([
      self.dom.deselect_tabs(),
      tab.dom.setAttribute("data-selected", True).r,
      self._report.js.getElementsByName(self.panels_name).all([
        #self._report.js.getElementsByName(self.tabs_name).all([
        #  self._report.js.data.all.element.css(self.options.tab_not_clicked_style(name))]),
        tab.dom.css(tab_style_clicked),
        self._report.js.data.all.element.hide(),
        tab_container.dom.toggleClass(css_cls_name, propagate=True) if css_cls_name is not None else "",
        div.dom.show()])])
    tab.inReport = False
    self.__panel_objs[name] = {"tab": tab_container, "content": div}
    if selected:
      self.__selected = name
    return self

  def __str__(self):
    if self.__selected is not None:
      self.__panel_objs[self.__selected]["content"].style.css.display = 'block'
      self.__panel_objs[self.__selected]["tab"][0].css(self.options.tab_clicked_style(self.__selected))
      self.__panel_objs[self.__selected]["tab"][0].attr["data-selected"] = 'true'
    content = []
    for p in self.__panels:
      self.tabs_container += self.__panel_objs[p]["tab"]
      content.append(self.__panel_objs[p]["content"].html())
    return "<div %s>%s%s</div>%s" % (self.get_attrs(pyClassNames=self.style.get_classes()), self.tabs_container.html(), "".join(content), self.helper)


class TabsArrowsDown(Tabs):
  name, category, callFnc = 'Tabs', 'Layouts', 'tabs'

  def add_panel(self, name, div, icon=None, selected=False, css_tab=None, css_tab_clicked=None):
    super(TabsArrowsDown, self).add_panel(name, div, icon, selected, css_tab, css_tab_clicked)
    self.tab_holder(name).style.add_classes.layout.panel_arrow_down()
    return self


class TabsArrowsUp(Tabs):
  name, category, callFnc = 'Tabs', 'Layouts', 'tabs'

  def add_panel(self, name, div, icon=None, selected=False, css_tab=None, css_tab_clicked=None):
    super(TabsArrowsUp, self).add_panel(name, div, icon, selected, css_tab, css_tab_clicked)
    self.tab_holder(name).style.add_classes.layout.panel_arrow_up()
    return self


class IFrame(Html.Html):
  name, category, callFnc = 'IFrame', 'Container', 'iframe'

  def __init__(self, report, url, width, height, helper, profile):
    super(IFrame, self).__init__(report, url, css_attrs={"width": width, "height": height}, profile=profile)
    self.css({"overflow-x": 'hidden'})
    self.add_helper(helper)

  @property
  def _js__builder__(self):
    return 'htmlObj.src = data'

  def __str__(self):
    return "<iframe src='%s' %s frameborder='0' scrolling='no'></iframe>%s" % (self.val, self.get_attrs(pyClassNames=self.style.get_classes()), self.helper)


class Dialog(Html.Html):
  name, category, callFnc = 'DialogMenu', 'Layouts', 'dialogs'
  __reqCss, __reqJs = ['jqueryui'], ['jqueryui']

  def __init__(self, report, recordSet, width, height, helper, profile):
    super(Dialog, self).__init__(report, recordSet, css_attrs={"width": width, 'height': helper}, profile=profile)
    self.css({"border": '2px solid %s' % self._report.theme.greys[3], "display": "block", "position": "absolute",
              "background": self._report.theme.greys[0]})
    # self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.dom.jquery_ui.draggable().toStr())

  @property
  def _js__builder__(self):
    return '''
      var div = jQuery("<div>")
      jQuery(htmlObj).append(div);
      div.dialog({modal: false, title: "rrrr"})'''

  # def jsAdd(self, title='data.event_val', isPyData=False):
  #   if isPyData:
  #     title = json.loads(title)
  #
  #   return '''
  #   var dialogWindow = $("<div>");
  #   var table = $("table");
  #   dialogWindow.append(table);
  #   dialogWindow.append('<input type="text">');
  #   var d = dialogWindow.dialog( { modal: false, title: %(title)s, show: 'puff', fluid: true,
  #       close: function () {$(this).remove()}, appendTo: "#%(htmlId)s", resizable: false,
  #       buttons: [{text: "Close", click: function() { $( this ).dialog("close")} } ]
  #   });
  #   d.parent().draggable({containment: '#%(htmlId)s'});
  #   event.preventDefault()''' % {'title': title, 'htmlId': self.htmlId}

  def __str__(self):
    return "<div %s></div>" % self.get_attrs(pyClassNames=self.style.get_classes())


class IconsMenu(Html.Html):
  name, category, callFnc = 'Icons Menu', 'Layouts', 'menu'
  __reqCss, __reqJs = ['font-awesome'], ['font-awesome']

  def __init__(self, icon_names, report, width, height, htmlCode, helper, profile):
    super(IconsMenu, self).__init__(report, None, width=width, css_attrs={"width": width, "height": height}, code=htmlCode,
                                    profile=profile)
    self._jsActions, self._definedActions = {}, []
    self._icons, self.icon = [], None
    self.css({"margin": "5px 0"})
    for i in icon_names:
      self.add_icon(i)

  def __getitem__(self, i):
    return self._icons[i]

  def add_icon(self, text, css=None, position="after"):
    """
    Add an icon to the HTML object

    Example
    checks.title.add_icon("fas fa-align-center")

    Documentation

    :param text: The icon reference from font awsome website
    :param css: Optional. A dictionary with the CSS style to be added to the component
    :param position:
    :return: The Html object
    """
    if text is not None:
      self._icons.append(self._report.ui.images.icon(text).css({"margin-right": '5px', 'cursor': "pointer"}))
      self.icon = self._icons[-1]
      if position == "before":
        self.prepend_child(self.icon)
      else:
        self.append_child(self.icon)
      #elf.icon.inReport = False
      if css is not None:
        self.icon.css(css)
    return self

  def add_select(self, action, data, width=150):
    options = ["<option>%s</option>" % d for d in data]
    self._jsActions[action] = '<select id="inputState" class="form-control" style="width:%spx;display:inline-block">%s</select>' % (width, "".join(options))
    self._definedActions.append(action)
    return self

  def __str__(self):
    htmlIcons = [htmlDef for action, htmlDef in self._jsActions.items()]
    return "<div %s>%s</div>" % (self.get_attrs(pyClassNames=self.style.get_classes()), "".join(htmlIcons))


class Form(Html.Html):
  name, category, callFnc = 'Generic Form', 'Forms', 'form'

  def __init__(self, report, htmlObjs, helper):
    super(Form, self).__init__(report, [])
    self.style.css.padding = "5px"
    self.add_helper(helper)
    self.__submit, self._has_container = None, False
    for i, htmlObj in enumerate(htmlObjs):
      self.__add__(htmlObj)

  def __getitem__(self, i):
    """
    Return the internal column in the row for the given index

    :param i: the column index
    :rtype: Col
    """
    return self.val[i]

  def __add__(self, htmlObj):
    """ Add items to a container """
    htmlObj.inReport = False # Has to be defined here otherwise it is set too late
    htmlObj.css({'text-align': 'left'})
    self.val.append(htmlObj)
    return self

  def submit(self, method, action="#", text="Submit"):
    if action is not None and method is not None:
      self.attr.update({"action": action, "method": method})

    self.__submit = self._report.ui.button(text).set_attrs({"type": 'submit'})
    self.__submit.inReport = False
    if self._has_container:
      self[0] + self.__submit
    return self

  def __str__(self):
    if self.__submit is None:
      raise Exception("Submit must be defined in a form ")

    str_vals = "".join([i.html() for i in self.val]) if self.val is not None else ""
    return '<form %s>%s%s</form>%s' % (self.get_attrs(pyClassNames=self.style.get_classes()), str_vals, self.__submit.html(), self.helper)


class Modal(Html.Html):
  name, category, callFnc = 'Modal Popup',  'Container', 'modal'

  def __init__(self, report, htmlObjs, header, footer, submit, helper):
    """
    Description:
    -----------
    Constructor for the modal item.
    This object is composed of three parts:  a header, which is a row of object, a body which is a column and a footer which is a row
    they all accept collections of html objects and are configurable just like the normal rows and column objects

    """
    super(Modal, self).__init__(report, [])
    self.add_helper(helper)
    self.doSubmit = submit
    if self.doSubmit:
      self.submit = report.ui.buttons.important("Submit").set_attrs({"type": 'submit'})
      self.submit.inReport = False
    self.closeBtn = report.ui.texts.span('&times', width='auto')
    self.closeBtn.css(None, reset=True)
    self.closeBtn.style.add_classes.div.span_close()
    self.closeBtn.click(report.js.getElementById(self.htmlId).css({'display': "none"}))
    self.__header = report.ui.row([])
    self.__header.inReport = False
    if header:
      for obj in header:
        self.__header + obj
    self.__header += self.closeBtn
    if footer:
      for obj in footer:
        self.__footer + obj
    self.__footer = report.ui.row([])
    self.__footer.inReport = False
    self.__body = report.ui.col([]).css({'position': 'relative',  'overflow-y': 'scroll'})
    self.__body.inReport = False
    self.col = report.ui.col([self.__header, self.__body, self.__footer]).css({'width': 'auto'}, reset=True)
    self.col.style.add_classes.div.modal_content()
    self.col.inReport = False
    self.val.append(self.col)
    self.__outOfScopeClose = True
    for htmlObj in htmlObjs:
      self.__add__(htmlObj)

  @property
  def outOfScopeClose(self):
    return self.__outOfScopeClose

  @outOfScopeClose.setter
  def outOfScopeClose(self, val):
    self.__outOfScopeClose = val

  @property
  def style(self):
    """
    Property to the CSS Style of the component

    :rtype: GrpClsButton.ClassButton
    """
    if self._styleObj is None:
      self._styleObj = GrpClsContainer.ClassModal(self)
    return self._styleObj

  @property
  def header(self):
    return self.__header

  @property
  def footer(self):
    return self.__footer

  @property
  def body(self):
    return self.__body

  def show(self):
    return self._report.js.getElementById(self.htmlId).css({'display': 'block'})

  def close(self):
    return self._report.js.getElementById(self.htmlId).css({'display': 'none'})

  def close_on_background(self):
    """
    Description:
    ------------
    Will allow an event to close the modal if a click event is detected anywhere outside the modal
    """
    modal = self._report.js.getElementById(self.htmlId)
    self._report.js.addOnReady(self._report.js.window.events.
                                      addClickListener(self._report.js.if_('event.target == %s' % modal, modal.css({'display': 'none'})),
                                                       subEvents=['event']))

  def __add__(self, htmlObj):
    """ Add items to a container """
    htmlObj.inReport = False # Has to be defined here otherwise it is set too late
    self.__body += htmlObj
    return self

  def __str__(self):
    if self.__outOfScopeClose:
      self.close_on_background()
    str_vals = "".join([i.html() for i in self.val]) if self.val is not None else ""
    self.set_attrs({'css': self.style.css.attrs})
    if self.doSubmit:
      self.col += self.submit
    return '<div %s>%s</div>%s' % (self.get_attrs(pyClassNames=self.style.get_classes()), str_vals, self.helper)


class Indices(Html.Html):
  name, category, callFnc = 'Index', 'Panels', 'index'
  __reqCss, __reqJs = ['font-awesome'], ['font-awesome']

  def __init__(self, report, count, width, height, htmlCode, options, profile):
    super(Indices, self).__init__(report, count, css_attrs={"width": width, "height": height}, profile=profile)
    self.items = []
    self.__options = OptPanel.OptionsPanelPoints(report, options)
    for i in range(count):
      div = self._report.ui.div(i, width=(15, "px"))
      div.attr["name"] = self.htmlId
      div.attr["data-position"] = i + 1
      div.css({"display": 'inline-block', "padding": "2px", "text-align": "center"})
      div.css(self.options.div_css)
      div.style.add_classes.div.background_hover()
      div.inReport = False
      self.items.append(div)
    #
    self.first = self._report.ui.icon("fas fa-angle-double-left", width=(20, 'px')).css({"display": 'inline-block'})
    self.first.inReport = False
    self.prev = self._report.ui.icon("fas fa-chevron-left", width=(20, 'px')).css({"display": 'inline-block'})
    self.prev.inReport = False
    self.next = self._report.ui.icon("fas fa-chevron-right", width=(20, 'px')).css({"display": 'inline-block'})
    self.next.inReport = False
    self.last = self._report.ui.icon("fas fa-angle-double-right", width=(20, 'px')).css({"display": 'inline-block'})
    self.last.inReport = False

  @property
  def options(self):
    """

    :rtype: OptPanel.OptionsPanelPoints
    """
    return self.__options

  def __getitem__(self, i):
    return self.items[i]

  def click_item(self, i, jsFncs, profile=False):
    """
    Description:
    ------------

    :param i:
    :param jsFncs:
    :param profile:
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    return self[i].on("click", [
      self[i].dom.by_name.css({"border-bottom": "1px solid %s" % self._report.theme.colors[0]}).r,
      self[i].dom.css({"border-bottom": "1px solid %s" % self.options.background_color})] + jsFncs, profile)

  def __str__(self):
    str_vals = "".join([self.first.html(), self.prev.html()] + [i.html() for i in self.items] + [self.next.html(), self.last.html()])
    return '<div %s>%s</div>%s' % (self.get_attrs(pyClassNames=self.style.get_classes()), str_vals, self.helper)


class Points(Html.Html):
  name, category, callFnc = 'Index', 'Panels', 'index'

  def __init__(self, report, count, width, height, htmlCode, options, profile):
    super(Points, self).__init__(report, count, css_attrs={"width": width, "height": height}, profile=profile)
    self.items = []
    self.css({"text-align": "center"})
    self.__options = OptPanel.OptionsPanelPoints(report, options)
    for i in range(count):
      div = self._report.ui.div(self._report.entities.non_breaking_space)
      div.attr["name"] = self.htmlId
      div.attr["data-position"] = i # keep the python indexation
      div.css({"border": "1px solid %s" % self._report.theme.greys[5], "border-radius": "10px", "width": "15px", "height": "15px"})
      div.css(self.options.div_css)
      div.style.add_classes.div.background_hover()
      div.inReport = False
      self.items.append(div)
    self[self.options.selected].css({"background-color": self.options.background_color})

  @property
  def options(self):
    """
    Description:
    ------------

    :rtype: OptPanel.OptionsPanelPoints
    """
    return self.__options

  def on(self, event, jsFncs, profile=False):
    """
    Description:
    ------------
    Add Javascript events to all the items in the component

    Attributes:
    ----------
    :param jsFncs: Array. The Javascript functions
    :param profile:
    :return:
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    str_fnc = JsUtils.jsConvertFncs(jsFncs, toStr=True)
    if event == "click":
      for i in range(len(self.items)):
        self.click_item(i, str_fnc)
    else:
      for i in range(len(self.items)):
        self.on_item(i, event, str_fnc)
    return self

  def on_item(self, i, event, jsFncs, profile=False):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param i: Integer. The item index
    :param event: String. The Javascript event reference
    :param jsFncs: Array. The Javascript functions
    :param profile:
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    return self[i].on(event, [
      'var data = {position: this.getAttribute("data-position")}'] + jsFncs, profile)

  def click_item(self, i, jsFncs, profile=False):
    """
    Description:
    ------------
    Add a click event on a particular item of the component

    Attributes:
    ----------
    :param i: Integer. The item index
    :param jsFncs: Array. The Javascript functions
    :param profile:
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    return self[i].on("click", [
      'var data = {position: this.getAttribute("data-position")}',
      self[i].dom.by_name.css({"background-color": ""}).r,
      self[i].dom.css({"background-color": self.options.background_color})] + jsFncs, profile)

  def __getitem__(self, i):
    return self.items[i]

  def __str__(self):
    str_vals = "".join([i.html() for i in self.items])
    return '<div %s>%s</div>%s' % (self.get_attrs(pyClassNames=self.style.get_classes()), str_vals, self.helper)


class Header(Html.Html):

  def __init__(self, report, htmlObj, width, height, htmlCode, helper, options, profile):
    super(Header, self).__init__(report, htmlObj, code=htmlCode, css_attrs={"width": width, "height": height}, profile=profile)
    self.__options = OptPanel.OptionsDiv(self, options)
    self.add_helper(helper)

  @property
  def options(self):
    """
    Description:
    ------------
    Property to set all the possible object for a button

    :rtype: OptPanel.OptionsDiv
    """
    return self.__options

  def __add__(self, htmlObj):
    """ Add items to a container """
    htmlObj.inReport = False # Has to be defined here otherwise it is set to late
    if self.options.inline:
      htmlObj.style.css.display = 'inline-block'
    self.val.append(htmlObj)
    return self

  def __getitem__(self, i):
    return self.val[i]

  def __str__(self):
    str_div = "".join([v.html() if hasattr(v, 'html') else str(v) for v in self.val])
    return "<header %s>%s</header>%s" % (self.get_attrs(pyClassNames=self.style.get_classes()), str_div, self.helper)


class Section(Html.Html):

  def __init__(self, report, htmlObj, width, height, htmlCode, helper, options, profile):
    super(Section, self).__init__(report, htmlObj, code=htmlCode, css_attrs={"width": width, "height": height}, profile=profile)
    self.__options = OptPanel.OptionsDiv(self, options)
    self.add_helper(helper)

  @property
  def options(self):
    """
    Description:
    ------------
    Property to set all the possible object for a button

    :rtype: OptPanel.OptionsDiv
    """
    return self.__options

  def __add__(self, htmlObj):
    """ Add items to a container """
    htmlObj.inReport = False # Has to be defined here otherwise it is set to late
    if self.options.inline:
      htmlObj.style.css.display = 'inline-block'
    self.val.append(htmlObj)
    return self

  def __getitem__(self, i):
    return self.val[i]

  def __str__(self):
    str_div = "".join([v.html() if hasattr(v, 'html') else str(v) for v in self.val])
    return "<section %s>%s</section>%s" % (self.get_attrs(pyClassNames=self.style.get_classes()), str_div, self.helper)




"""

"""

from epyk.core.css.catalogs import Catalog

from epyk.core.css.styles.classes import CssStylesRadio


class CatalogRadio(Catalog.CatalogGroup):
  def button(self):
    """  """
    return self._set_class(CssStylesRadio.CssRadioButton)

  def selected(self):
    """  """
    return self._set_class(CssStylesRadio.CssRadioButtonSelected)

  def switch_checkbox(self):
    """  """
    return self._set_class(CssStylesRadio.CssRadioSwitch)

  def switch_label(self):
    """  """
    return self._set_class(CssStylesRadio.CssRadioSwitchLabel)

  def switch_checked(self):
    """  """
    return self._set_class(CssStylesRadio.CssRadioSwitchChecked)



"""
Base Class to create a database and perform SQL operations using the sqlalchemy interface

Related Pages:

			https://www.pythonsheets.com/notes/python-sqlalchemy.html#join-joined-two-tables-via-join-statement

"""

import json
import inspect
import urllib
import importlib
import logging
import sys
import os
import datetime
import traceback

# Check if pandas is available in the current environment
# if it is the case this module can handle Dataframe directly
try:
  import pandas as pd
  has_pandas = True

except:
  has_pandas = False

from epyk.core.js.Imports import requires

# Will automatically add the external library to be able to use this module
sqlalchemy = requires("sqlalchemy", reason='Missing Package', install='sqlalchemy', source_script=__file__)

DATABASE_AUTOSYNC = False


class SqlConn(object):
  def __init__(self, family, database=None, filename=None, model_path=None, reset=False, migrate=True, tables_scope=None, **kwargs):
    """
    Here we try to setup as generic as we can all the variable environment variables for the DB.
    We try to not rely on the rptObj in order to be able to use this interface for various usage
    """
    import sqlalchemy.orm

    self.query = None  # In this design we decided to go to a route where each user will manage his database and also the person who can look at the data
    # This will simplify a lot the permissionning (we consider the storage not a problem at this stage as we split per user
    # Also this could be interesting to check the user data use
    # Some reports can get centralised databases using the module variable SINGLEDB
    dbConfig = {'drivername': kwargs["driver"] if "driver" in kwargs else family, 'username': kwargs.get('username'),
                'password': kwargs.get('password'), 'host': kwargs.get('host'),
                'port': kwargs.get('port'), 'query': kwargs.get('query'), 'database': database}
    self.dbPath = database
    self.username = kwargs.get('username')
    self.userhost = kwargs.get('userhost')
    if kwargs.get('driver', '') == "mssql+pyodbc":
      # Special logic to create a connection to a MS SQl database
      # https://www.linkedin.com/pulse/connecting-sql-database-python-patrick-j-ryan
      try:
        params = urllib.parse.quote_plus('DRIVER=' + kwargs.get('driverName', "{ODBC Driver 17 for SQL Server}") + ';SERVER=' + kwargs['host'] + ';DATABASE=' + database + ';Trusted_Connection=yes;')
      except:
        params = urllib.quote_plus('DRIVER=' + kwargs.get('driverName', "{ODBC Driver 17 for SQL Server}") + ';SERVER=' + kwargs['host'] + ';DATABASE=' + database + ';Trusted_Connection=yes;')
      # echo is dedicated to keep track of changes and store those to logs files
      self.engine = sqlalchemy.create_engine('mssql+pyodbc:///?odbc_connect=%s' % params, echo=kwargs.get('echo', False))
    else:
      self.engine = sqlalchemy.create_engine(sqlalchemy.engine.url.URL(**dbConfig), echo=kwargs.get('echo', False))
    self.metadata = sqlalchemy.MetaData(bind=self.engine)
    self.session = sqlalchemy.orm.sessionmaker(bind=self.engine)()
    # The database reflection is only used in some special cases by default (internally for private en public db)
    if tables_scope is None:
      self.metadata.reflect(bind=self.engine)
    else:
      self.metadata.reflect(only=list(tables_scope), bind=self.engine)
    if kwargs.get('reflect', DATABASE_AUTOSYNC):
      self.metadata.create_all(self.engine)
      if model_path:
        self.load_schema(filename=filename, model_path=model_path, reset=reset)

  def _load_sql_file(self, filename, reset):
    """
    Load a database schema from a python file.
    This will use the dialect available from SQLAlchemy

    Related Pages:

			https://www.sqlalchemy.org/

    :param filename: The database schema python file name
    :param reset: Boolean, flag to specify is the schema needs to be fully reloaded

    :return:
    """
    on_init_fnc = None
    model_mod = importlib.import_module(filename.replace('.py', ''))
    for table_name, table in inspect.getmembers(model_mod):
      if table_name == 'on_init':
        on_init_fnc = table
      elif '__' not in table_name and inspect.isroutine(table) and table.__module__ in filename:
        table_def = getattr(model_mod, table_name)()
        table_def.append(sqlalchemy.Column('lst_mod_dt', sqlalchemy.DateTime, default=datetime.datetime.utcnow, nullable=True))
        if not self.engine.has_table(table_name):
          new_table = sqlalchemy.Table(table_name, self.metadata, *table_def)
          new_table.create(self.engine, checkfirst=True)
        else:
          # Compare the two schema and add the columns if necessary
          # TO DO: Find a way to drop a column without removing the data in the database
          new_cols = set([c.key for c in table_def])
          curr_cols = set([c.key for c in self.metadata.tables[table_name].columns])
          for cn in new_cols.difference(curr_cols):
            for c in table_def:
              if c.key == cn:
                if c.default is not None:
                  self.engine.execute('ALTER TABLE %s ADD COLUMN %s %s DEFAULT %s' % (table_name, cn, c.type.compile(self.engine.dialect), json.dumps(c.default.arg)))
                else:
                  self.engine.execute('ALTER TABLE %s ADD COLUMN %s %s' % (table_name, cn, c.type.compile(self.engine.dialect)))
          if reset or table_name in getattr(model_mod, 'RESET_TABLES', []):
            #self.metadata.tables[tableName].drop(self.engine, checkfirst=False)
            new_table = sqlalchemy.Table(table_name, self.metadata, *table_def, extend_existing=True)
            new_table.drop(self.engine, checkfirst=True)
            if not table_name in getattr(model_mod, 'RESET_TABLES', []):
              new_table.create(self.engine, checkfirst=True)
    #We do the part where we run default that need to happen on database creation
    if on_init_fnc:
      on_init_fnc(self)

  def table(self, table_name):
    """
    Return a sqlAlchemy table object. This can be useful in the where clauses

    Example
    db.table('table1')

    :param table_name:
    :return: Python table object
    """
    if self.engine.has_table(table_name):
      return sqlalchemy.Table(table_name, self.metadata)

    raise Exception('Table %s does not exist in the schema' % table_name)

  def load_schema(self, filename=None, model_path=None, reset=False):
    """
    Function that takes care of initialising the DB
    Please note that some column names are prohibited such as lst_mod_dt

    Example

    :param filename: The python module used to get the database schema
    :param model_path: The python path with the model
    :param reset: Boolean, Flag to reset the database. THis will emtpy the tables
    :return:
    """
    if not filename and not model_path:
      raise Exception("You need to specify at least a file name or a model path")

    if model_path:
      sys.path.append(model_path)
      for pyFile in os.listdir(model_path):
        if not pyFile.endswith('.py') or pyFile == '__init__.py':
          continue

        if filename and filename != pyFile:
          continue

        self._load_sql_file(pyFile, reset)
    elif filename:
      self._load_sql_file(filename, reset)

  def table_clone(self, old_table, new_table, mapping=None):
    """
    Helps to migrate between two tables.
    The mapping argument is used in case the column names differ between the two tables

    :param old_table: A SQLAlchemy table class in the current database model
    :param new_table: A SQLAlchemy table class
    :param mapping: Optional, A dictionary for the column names

    :return:
    """
    old_schema = sqlalchemy.Table(old_table, self.metadata)
    new_schema = sqlalchemy.Table(new_table, self.metadata)
    if self.engine.has_table(new_table):
      new_schema.drop(self.engine, checkfirst=True)
    for column in old_schema.columns:
      column.name = mapping.get(column.name, column.name) if mapping else column.name
      new_schema.append_column(column)
    new_schema.create(self.engine, checkfirst=True)
    return self

  def table_migrate(self, from_table, to_table):
    """
    Copy data from one table to another

    :param from_table:
    :param to_table:

    :return: The Python SQL object
    """
    data = list(self.select([from_table]).data())
    self.insert(to_table, data.to_dict('records'), commit=True)
    return self

  def table_create(self, table_name, table_def, reset=False):
    """

    Example
      db = rptObj.db(database=r"newTest.db")
      tableDef = [sqlalchemy.Column('environment', sqlalchemy.String, nullable=False),
                  sqlalchemy.Column('report', sqlalchemy.String, nullable=False),]
      db.createTable('newTable', tableDef)

    :param table_name:
    :param table_def:
    :param reset:

    """
    if not self.engine.has_table(table_name):
      new_table = sqlalchemy.Table(table_name, self.metadata, *table_def)
      new_table.create(self.engine, checkfirst=True)
    else:
      if reset:
        new_table = sqlalchemy.Table(table_name, self.metadata, *table_def)
        new_table.drop(self.engine, checkfirst=True)
        new_table.create(self.engine, checkfirst=True)
    return self

  def table_empty(self, table_name):
    """
    This function will empty an existing table

    Example
    db.emptyTable('test')

    :param table_name: A string with the datable name

    :return: self
    """
    self.delete(table_name)
    logging.info('Content of table %s deleted' % table_name)
    self.commit()
    return self

  def table_create_from_file(self, filename, table_name, records=None, path=None, reset=False, commit=True):
    """

    Example
    df = rptObj.file(htmlCode=r"IBRD_Balance_Sheet__FY2010.csv").read()
    db = rptObj.db(database=r"newTest.db").forceCreate()
    dbObj.createTable('myschema.py', 'mytable', records=df)

    :param filename:
    :param table_name:
    :param records:
    :param path:
    :param reset:
    :param commit:

    :return: The Python SQL object
    """
    self.load_schema(model_path=path, filename=filename, reset=reset)
    if records:
      self.insert(records=records.records(), commit=commit, table_name=table_name)
    return self

  def force_create(self):
    """
    Force the creation of the database in the given project

    :return: The python Sql object
    """
    from sqlalchemy.ext.declarative import declarative_base

    declarative_base().metadata.create_all(self.engine)
    return self

  def load_data_file(self, filename, path, reset=False, new_tables=None):
    """
    Load a python sql file to the local database.
    This will only add records and then commit the changes.

    Those data should not be sensitive ones if they are store and committed to the folder.

    :param filename:
    :param path:
    :param reset:
    :param new_tables:

    :return: The Python SQL object
    """
    sys.path.append(path)
    py_mod = importlib.import_module(filename.replace(".py", ""))
    if hasattr(py_mod, 'data') and hasattr(py_mod, 'target'):
      conn = self.engine.connect()
      header = py_mod.data[0]
      sql_target = self.table(py_mod.target)
      if reset:
        conn.execute(sql_target.delete())
      if py_mod.target in new_tables or new_tables is None:
        if isinstance(header, list):
          for rec in py_mod.data[1:]:
            conn.execute(sql_target.insert().values(dict(zip(header, rec))))
        else:
          for rec in py_mod.data:
            conn.execute(sql_target.insert().values(rec))
    return self

  def where(self, stmts):
    """
    Add a where clause to the SqlAlchemy query.

    Examples
    db.select().where([db.column("table", 'column') == 'X')
    db.select( ['BNP'] ).where([ db.column('BNP', 'date') == '22/04/2013 00:00'] ).toDf()

    :param stmts: The SQL where statement

    :return: The python object itself
    """
    for stmt in stmts:
      self.query = self.query.where(stmt)
    return self

  def select(self, table_name=None, columns=None):
    """
    Create a SQL statement

    Example
    rptObj.db().select(["worldcup_teams"])

    Documentation
    http://docs.sqlalchemy.org/en/latest/core/selectable.html
    http://docs.sqlalchemy.org/en/latest/core/sqlelement.html

    :param table_name: Optional, The database table name
    :param columns: Optional, The list of columns

    :return: self
    """
    if table_name is not None:
      if not isinstance(table_name, list):
        table_name = [table_name]
      tables = [sqlalchemy.Table(table, self.metadata, autoload=True) for table in table_name]
      self.query = sqlalchemy.sql.select(tables)
    elif columns:
      self.query = sqlalchemy.sql.select(columns)
    return self

  def delete(self, table_name):
    """
    Create a SQL delete SQL statement

    Example
    rptObj.db().delete('table1')

    Related Pages:

			https://docs.sqlalchemy.org/en/13/core/dml.html

    :param table_name: Optional, The database table name
    :return: self
    """
    self.query = sqlalchemy.Table(table_name, self.metadata, autoload=True).delete()
    return self

  def or_(self, *args):
    return sqlalchemy.sql.or_(*args)

  def and_(self, *args):
    return sqlalchemy.sql.and_(*args)

  def order_by(self, col):
    self.query = self.query.order_by(col)
    return self

  def desc(self, val):
    return sqlalchemy.desc(val)

  def asc(self, val):
    return sqlalchemy.asc(val)

  def update(self, table_name, values):
    """
    Create a delete SQL statment

    Example
    rptObj.db().update('table1', {db.column('test', 'name'): 'user'})

    Documentation
    http://docs.sqlalchemy.org/en/latest/core/selectable.html
    http://docs.sqlalchemy.org/en/latest/core/sqlelement.html
    https://docs.sqlalchemy.org/en/13/core/dml.html

    :param table_name:
    :param values:
    :return: self
    """
    self.query = sqlalchemy.Table(table_name, self.metadata, autoload=True).update(values=values)
    return self

  def distinct(self, columns=None):
    """

    :param columns:
    :return:
    """
    if columns is None:
      self.query = self.query.distinct()
    else:
      self.query = self.query.distinct(columns)
    return self

  def get_last_id(self, table_name):
    """
    Return the table last primary key ID.
    This will return an error if the table does not have a primary key defined in its schema

    Example
    db.get_last_id("table")

    :param table_name: The table name
    :return: Returh the last row ID or -1
    """
    table = sqlalchemy.Table(table_name, self.metadata)
    try:
      row = self.session.query(table).order_by(list(table.primary_key.columns)[0].desc()).first()
    except:
      raise Exception("No primary key defined for the table %s - db: %s" % (table_name, self.dbPath))

    if row is None:
      return -1

    rec = dict(zip(row.keys(), row))
    return rec[list(table.primary_key.columns)[0].key]

  def insert(self, table_name, records, commit=False, col_user_name=None, clean_rec=False, getIdCol=False):
    """
    insert a list of records to a table

    Example
    db.insert('table1',[{'name': 'test'}], commit=True)
    db.insert("table", {"user_name": "Test", "data": "test"}, commit=True)

    Related Pages:

			https://docs.sqlalchemy.org/en/13/core/dml.html
    https://docs.sqlalchemy.org/en/13/core/dml.html

    :param table_name: The database table name
    :param records: The list of dictionaries with the data to inserts
    :param commit: Optional, Boolean to commit the insert. Set to False by default
    :param col_user_name:
    :param clean_rec: Optional, remove the key in the dictionaries which are not related to the table. Set to False
    :param getIdCol:

    :return: The python object itself
    """
    dflt = {'lst_mod_dt': datetime.datetime.utcnow()}
    error_count, error_log, last_id = 0, [], None
    table = sqlalchemy.Table(table_name, self.metadata)
    if not self.engine.has_table(table.name):
      raise Exception("Table %s does not exist" % table.name)

    if col_user_name is not None:
      dflt[col_user_name] = self.username
    if 'hostname' in self.columns(table_name):
      dflt['hostname'] = self.userhost
    if isinstance(records, dict):
      records = [records]
    if clean_rec:
      # This is done to remove the extra columns in the rec before creating the SQL statement
      tableCols = [str(col).split(".")[1] for col in table.columns]
      for rec in records:
        try:
          row = {}
          for c in tableCols:
            if c in rec:
              row[c] = rec[c]
          row.update(dflt)
          result = self.session.execute(table.insert().values(row))
          last_id = result.inserted_primary_key
        except Exception as err:
          logging.warning(traceback.format_exc())
          error_count += 1
          error_log.append(str(traceback.format_exc()).strip().split('\n')[-1])
    else:
      for rec in records:
        try:
          tmpRec = dict(rec)
          tmpRec.update(dflt)
          result = self.session.execute(table.insert().values(tmpRec))
          last_id = result.inserted_primary_key
        except Exception as err:
          logging.warning(traceback.format_exc())
          error_count += 1
          error_log.append(str(traceback.format_exc()).strip().split('\n')[-1])
    if commit:
      self.session.commit()
    if error_count:
      return (False, error_count, error_log, last_id)

    return (True, 0, [], last_id)

  def data(self, limit=None):
    """
    Returns the results of the select statement previously instantiated in a pandas dataframe

    Example
    rptObj.db().getData()

    :param limit: Optional. The number of records to be returned

    :return: A pandas dataframe
    """
    if self.query is None:
      return None

    try:
      import pandas

      if limit:
        return pandas.read_sql(self.query, self.query.bind).head(limit)

      return pandas.read_sql(self.query, self.query.bind)
    except:
      # No pandas installed
      return list(self.execute())

  @property
  def records(self):
    """
    Return the records

    Example
    for rec in db.select("table").records:
      print(rec)

    :return: A iterator for the SQL result
    """
    if has_pandas:
      for rec in self.data().to_dict('records'):
        yield rec

    else:
      for rec in self.data():
        yield dict(rec.items())

  @property
  def count(self):
    """
    Return the number of records

    Example
    print(db.select("table").count)

    :return: The number of records
    """
    results = self.data()
    if has_pandas:
      return results.count

    return len(results)

  def limit(self, n):
    """
    Limit the number of records returned

    :param n: Integer, the number of records
    :return: The SQL Query object
    """
    if self.query is None:
      return None

    self.query = self.query.limit(n)
    return self

  def first(self, items=False):
    """
    Return only the first items from the SQL query

    Example
    print(db.first())
    print(db.first(items=True))

    :param items: Return a dictionary or a list of data
    :return: None or the first record
    """
    self.limit(1)
    try:
      if items:
        row = self.engine.connect().execute(self.query).next()
        return dict(row.items())

      return self.engine.connect().execute(self.query).next()

    except StopIteration:
      return None

  @property
  def tables(self):
    """
    Return the list of tables defined in the selected database

    Example
    rptObj.db().tables()

    :return: A python object with the list of tables
    """
    return self.engine.table_names()

  def columns(self, table_name):
    """
    Return the list of columns defined in the selected database

    Example
    rptObj.db().columns("table_name")

    :param table_name:
    :return: A python object with the list of tables
    """
    table = sqlalchemy.Table(table_name, self.metadata)
    if self.engine.has_table(table.name):
      return table.columns

    raise Exception('Table does not exist')

  def column(self, table_name, column_name):
    """
    Return a sqlAlchemy column object. This can be useful in the where clauses

    Example
    select('table').where([db.column("table", 'column') == 'X')

    :param table_name: The database table name
    :param column_name: The column name
    :return: Python column object
    """
    table = sqlalchemy.Table(table_name, self.metadata, autoload=True)
    if self.engine.has_table(table.name):
      return getattr(table.c, column_name)

  def drop(self, table_name, validate=True):
    """
    Delete the table from the database.
    The pre check can be disabled and the table will be automatically created again when the report will be triggered again.
    No extra function to create a table in the framework this is done by the SQL framework itself

    Example
    rptObj.db().drop('test')
    rptObj.db().drop('test', withCheck=False)

    Related Pages:

			https://docs.sqlalchemy.org/en/13/core/connections.html#sqlalchemy.engine.ResultProxy

    :param table_name: The database table name
    :param validate: Boolean, Validation check before dropping the table
    :return: Python column object
    """
    if validate:
      try:
        name = input("Are you sure to delete the table %s (Y/N)? " % table_name)
      except:
        name = raw_input("Are you sure to delete the table %s (Y/N)? " % table_name)
      if name == 'Y':
        sqlalchemy.Table(table_name, self.metadata, autoload=True).drop()
        logging.info("Table %s deleted" % table_name)
    else:
      sqlalchemy.Table(table_name, self.metadata, autoload=True).drop()
      logging.info("Table %s deleted" % table_name)

  def execute(self):
    """
    Execute the current SQL query

    Related Pages:

			https://docs.sqlalchemy.org/en/13/core/connections.html#sqlalchemy.engine.ResultProxy

    :return: The SQL Result proxy
    """
    return self.engine.execute(self.query)

  def commit(self):
    """
    Commit the current transaction.

    THis will save the results in the database

    :return: The SQL object
    """
    self.session.commit()
    return self


class SqlConnOdbc(object):
  """
  Connector to Access databases. This connector will allow you to create, store and retrieve data from any MS Access Database.
  This will return the SQL database object. It will be possible to reuse the same syntax to then interact with it.

  This would need the ODBC driver available here: https://www.microsoft.com/en-us/download/confirmation.aspx?id=13255
  """

  def __init__(self, database=None, **kwargs):
    self.query = {}
    pyodbc = requires("pyodbc", reason='Missing Package', install='sqlalchemy', source_script=__file__)
    dbAttr = {'Driver': kwargs['driver'], 'DBQ': database}
    for opt in [('password', 'PWD'), ('username', 'UID')]:
      if kwargs.get(opt[0]) is not None:
        dbAttr[opt[1]] = kwargs[opt[0]]
    dbStr = []
    for k in ['Driver', 'DBQ', 'PWD', 'UID']:
      if k in dbAttr:
        dbStr.append("%s=%s" % (k, dbAttr[k]))
    self.conn = pyodbc.connect(";".join(dbStr))

  def select(self, tableNames):
    if not isinstance(tableNames, list):
      tableNames = [tableNames]
    self.query = {'tables': tableNames}
    return self

  def columns(self, tableName, columnName):
    if 'columns' not in self.query:
      self.query['columns'] = []
    self.query['columns'].append(self.column(tableName, columnName))
    return self

  def where(self, stmts):
    self.query['wheres'] = []
    for stmt in stmts:
      self.query['wheres'].append(stmt)
    return self

  def column(self, tableName, columnName):
    return "%s.%s" % (tableName, columnName)

  @property
  def sql(self):
    strSql = "SELECT %s FROM %s" % (", ".join(self.query.get('columns', ['*'])), ", ".join(self.query['tables']))
    if len(self.query.get('wheres', [])) > 0:
      strSql = "%s WHERE %s" % (strSql, " and ".join(self.query['tables']))
    return strSql

  def insert(self, tableName, records, commit=False, colUserName=None):
    for rec in records:
      row = "(%s" % json.dumps(rec[0]).replace('"', "'")
      for r in rec[1:]:
        row = "%s,%s" % (row, json.dumps(r).replace('"', "'"))
      row = "%s)" % row
      cursor = self.conn.execute('insert into %s values%s' % (tableName, row))
      if commit:
        cursor.commit()

  def data(self, limit=None):
    cursor = self.conn.cursor()
    return [rec for rec in cursor.execute(self.sql())]

  @property
  def tables(self):
    """

    :return:
    """
    cursor = self.conn.cursor()
    return [row.table_name for row in cursor.tables()]


class SqlConnNeo4j(object):
  def __init__(self, host=None, port=None, usr=None, pwd=None):
    import neo4j
    self.driver = neo4j.GraphDatabase.driver('bolt://%s:%s' % (host, port), auth=(usr, pwd))
    self.query = []

  def raw_query(self, query):
    try:
      with self.driver.session() as session:
        for rec in session.run(query):
          yield rec

    except:
      raise StopIteration

  def create(self):
    self.query.append('CREATE')
    return self

  def match(self):
    self.query.append('MATCH')
    return self

  def foreach(self, conditions):
    """

    """
    raise NotImplemented

  def where(self, condition):
    return self

  def delete(self, node_names, detach=False):

    if detach:
      self.query.append('DETACH')
    self.query.append('DELETE %s' % ','.join(node_names))
    try:
      with self.driver.session() as session:
        session.run(self.compose(self.query))
        self.query = []
        return True

    except:
      print(traceback.format_exc())
      self.query = []
      return False

  def retrieve(self, names):
    self.query.append('RETURN')
    self.query.append(','.join(names))
    return self

  def clear(self):
    """
    :dsc: Clears all nodes and edges from the Database
    """
    return self.match().node('n').delete(['n'], detach=True)

  def node(self, name='', labels=None, attr=None):
    """
    :dsc: adds the node patern to the query
    """
    if not labels:
      labels = []
    else:
      labels[0] = ':%s' % labels[0]
    if not attr:
      attr = ''
    else:
      tmp_attr_list = []
      for attr_key, attr_value in attr.items():
        tmp_attr_list.append('%s: "%s"' % (attr_key, attr_value))
      attr = '{%s}' % ','.join(tmp_attr_list)
    self.query.append("(%s%s %s)" % (name, ':'.join(labels), attr))
    return self

  def link(self, labels='', attr=None, direction="from"):
    """
    :dsc: adds the edge definition to the query
    """
    if direction == 'from':
      self.query.append('-[%s]->' % labels)
    else:
      self.query.append('<-[%s]-' % labels)
    return self

  def alias(self, aliases):
    """
    :dsc: defines a set of aliases that will appear as WITH a, b, c, d as count(id)
    The aliases argument will be defined as follows: ['a', 'b', 'c', {'d': 'count(id)'}]
    """
    self.query.append('WITH')
    tmp_query = []
    for expression in aliases:
      if isinstance(expression, dict):
        for expr, alias in expression.items():
          tmp_query.append('%s as %s' % (expr, alias))
      else:
        tmp_query.append(expression)
    self.query.append(', '.join(tmp_query))
    return self

  def compose(self, query):
    """
    :dsc: Simply joins the query clauses all together
    """
    return ' '.join(query)

  def execute(self):
    try:
      with self.driver.session() as session:
        for rec in session.run(self.compose(self.query)):
          yield rec

    except Exception as err:
      print(traceback.format_exc())
      raise StopIteration






"""
Wrapper to the Jquery UI package

Related Pages:

		https://api.jqueryui.com/
"""

import json

from epyk.core.js import JsUtils
from epyk.core.js.packages import JsPackage
from epyk.core.js.primitives import JsObjects


class JQueryUI(JsPackage):
  lib_alias = {"js": 'jqueryui', 'css': 'jqueryui'}
  lib_set_var = False

  def __init__(self, htmlObj, varName=None, setVar=True, isPyData=True, report=None):
    self.htmlId = varName if varName is not None else htmlObj.htmlId
    self.varName, self.varData, self.__var_def = "document.getElementById('%s')" % self.htmlId, "", None
    self._src, self._report = htmlObj, report
    self._js, self._jquery = [], None

  def labels(self):
    """
    Finds all label elements associated with the first selected element.

    :return:
    """
    self._js.append("labels()")
    return self

  def cssClip(self, css=None):
    """
    Getter/setter for an object version of the CSS clip property.

    Related Pages:

			https://api.jqueryui.com/cssClip/

    :param css:

    :return:
    """
    if css is not None:
      self._js.append("cssClip(%s)" % JsUtils.jsConvertData(css, None))
    else:
      self._js.append("cssClip()")
    return self

  def position(self, options=None):
    """
    The jQuery UI .position() method allows you to position an element relative to the window, document, another element, or the cursor/mouse, without worrying about offset parents.

    Related Pages:

			https://api.jqueryui.com/position/

    :param options:

    :return:
    """
    if options is not None:
      self._js.append("position(%s)" % JsUtils.jsConvertData(options, None))
    else:
      self._js.append("position()")
    return self

  def draggable(self, options=None):
    """
    Enable draggable functionality on any DOM element.
    Move the draggable object by clicking on it with the mouse and dragging it anywhere within the viewport.

    Example

    Related Pages:

			https://jqueryui.com/draggable/

    :param options:
    """
    if options is not None:
      return self.fnc("draggable(%s)" % JsUtils.jsConvertData(options, None))

    return self.fnc("draggable()")

  def droppable(self, options=None):
    """
    Enable any DOM element to be droppable, a target for draggable elements.

    Example

    Related Pages:

			https://jqueryui.com/droppable/

    :param options:

    :return:
    """
    if options is not None:
      self._js.append("droppable(%s)" % JsUtils.jsConvertData(options, None))
    else:
      self._js.append("droppable()")
    return self

  def sortable(self, options=None):
    """
    Enable a group of DOM elements to be sortable.
    Click on and drag an element to a new spot within the list, and the other items will adjust to fit.
    By default, sortable items share draggable properties.

    Related Pages:

			https://jqueryui.com/sortable/

    :param options:

    :return:
    """
    if options is not None:
      self._js.append("sortable(%s)" % JsUtils.jsConvertData(options, None))
    else:
      self._js.append("sortable()")
    return self

  def resizable(self, options=None):
    """
    Enable any DOM element to be resizable.
    With the cursor grab the right or bottom border and drag to the desired width or height.

    Example

    Related Pages:

			https://jqueryui.com/resizable/#constrain-area

    :param options:

    :return:
    """
    if options is not None:
      self._js.append("resizable(%s)" % JsUtils.jsConvertData(options, None))
    else:
      self._js.append("resizable()")
    return self

  def selectable(self, options=None):
    """
    Enable a DOM element (or group of elements) to be selectable.
    Draw a box with your cursor to select items. Hold down the Ctrl key to make multiple non-adjacent selections.

    Related Pages:

			https://jqueryui.com/selectable/

    :param options:

    :return:
    """
    if options is not None:
      self._js.append("selectable(%s)" % JsUtils.jsConvertData(options, None))
    else:
      self._js.append("selectable()")
    return self

  def addClass(self, cls_name, delay, callback):
    """
    Adds class(es) to elements while animating all style changes.

    Related Pages:

			https://jqueryui.com/addClass/

    :param cls_name:
    :param delay:
    :param callback:

    :return:
    """
    self._js.append("addClass('%s', %s, %s)" % (cls_name, delay, ";".join(JsUtils.jsConvertFncs(callback))))
    return self

  def removeClass(self, cls_name, delay, callback):
    """
    Removes class(es) from elements while animating all style changes.

    Related Pages:

			https://jqueryui.com/removeClass/

    :param cls_name:
    :param delay:
    :param callback:

    :return:
    """
    self._js.append("removeClass('%s', %s, %s)" % (cls_name, delay, ";".join(JsUtils.jsConvertFncs(callback))))
    return self

  def switchClass(self, cls_name, new_cls_name, delay):
    """
    Add and remove class(es) to elements while animating all style changes.

    :param cls_name:
    :param new_cls_name:
    :param delay:

    :return:
    """
    cls_name = JsUtils.jsConvertData(cls_name, None)
    new_cls_name = JsUtils.jsConvertData(new_cls_name, None)
    self._js.append("removeClass(%s, %s, %s)" % (cls_name, new_cls_name, delay))
    return self

  def toggleClass(self, cls_name, delay):
    """
    Toggle class(es) on elements while animating all style changes.

    Related Pages:

			https://jqueryui.com/toggleClass/

    :param cls_name:
    :param delay:

    :return:
    """
    cls_name = JsUtils.jsConvertData(cls_name, None)
    self._js.append("toggleClass(%s, %s)" % (cls_name, delay))
    return self

  def show(self, selected_effect=None, options=None, delay=None, callback=None):
    """
    Display elements using custom effects.

    Related Pages:

			https://jqueryui.com/show/

    :param selected_effect:
    :param options:
    :param delay:
    :param callback:

    :return:
    """
    if selected_effect is None:
      self._js.append("show()")
    else:
      selected_effect = JsUtils.jsConvertData(selected_effect, None)
      options = JsUtils.jsConvertData(options, None)
      callback = ";".join(JsUtils.jsConvertFncs(callback))
      self._js.append("show(%s, %s, %s, %s)" % (selected_effect, options, delay, callback))
    return self

  def hide(self, selected_effect=None, options=None, delay=None, callback=None):
    """
    Display elements using custom effects.

    Related Pages:

			https://jqueryui.com/show/

    :param selected_effect:
    :param options:
    :param delay:
    :param callback:

    :return:
    """
    if selected_effect is None:
      self._js.append("hide()")
    else:
      selected_effect = JsUtils.jsConvertData(selected_effect, None)
      options = JsUtils.jsConvertData(options, None)
      callback = ";".join(JsUtils.jsConvertFncs(callback))
      self._js.append("hide(%s, %s, %s, %s)" % (selected_effect, options, delay, callback))
    return self

  def toggle(self, selected_effect, options, delay):
    """

    Related Pages:

			https://jqueryui.com/toggle/

    :param selected_effect:
    :param options:
    :param delay:

    :return:
    """
    selected_effect = JsUtils.jsConvertData(selected_effect, None)
    options = JsUtils.jsConvertData(options, None)
    self._js.append("toggle(%s, %s, %s)" % (selected_effect, options, delay))
    return self

  def animate(self, css=None, delay=None):
    """
    Animate the properties of elements between colors.

    Related Pages:

			https://jqueryui.com/animate/

    :param css:
    :param delay:

    :return:
    """
    self._js.append("animate(%s, %s)" % (JsUtils.jsConvertData(css, None), delay))
    return self


class JQueryUiDatePicker(JQueryUI):

  def setDefaults(self):
    """

    :return:
    """

  def formatDate(self):
    """

    :return:
    """

  def option(self, name, value):
    """

    :return:
    """
    return "$('#%s input').datepicker( 'option', '%s', %s)" % (self.src.htmlId, name, json.dumps(value))

  def refresh(self):
    """

    :return:
    """


class Slider(JQueryUI):

  def destroy(self):
    """
    Description:
    ------------
    Removes the slider functionality completely. This will return the element back to its pre-init state.

    Related Pages:

			https://api.jqueryui.com/slider/#method-destroy
    """
    return JsObjects.JsObjects.get('%s.slider("destroy")' % self._src.dom.jquery.varId)

  def disable(self):
    """
    Description:
    ------------
    Disables the slider

    Related Pages:

			https://api.jqueryui.com/slider/#method-disable
    """
    return JsObjects.JsObjects.get('%s.slider("disable")' % self._src.dom.jquery.varId)

  def enable(self):
    """
    Description:
    ------------
    Enables the slider.

    Related Pages:

			https://api.jqueryui.com/slider/#method-enable
    """
    return JsObjects.JsObjects.get('%s.slider("enable")' % self._src.dom.jquery.varId)

  def instance(self):
    """
    Description:
    ------------
    Retrieves the slider's instance object. If the element does not have an associated instance, undefined is returned.

    Related Pages:

			https://api.jqueryui.com/slider/#method-instance
    """
    return JsObjects.JsObjects.get('%s.slider("instance")' % self._src.dom.jquery.varId)

  def option(self, jsData=None):
    """
    Description:
    ------------
    Retrieves the slider's instance object. If the element does not have an associated instance, undefined is returned.

    Related Pages:

			https://api.jqueryui.com/slider/#method-instance

    Attributes:
    ----------
    :param jsData:
    """
    if jsData is None:
      return JsObjects.JsObjects.get('%s.slider("option")' % self._src.dom.jquery.varId)

    jsData = JsUtils.jsConvertData(jsData, None)
    return JsObjects.JsObjects.get('%s.slider("option", %s)' % (self._src.dom.jquery.varId, jsData))

  def value(self, jsData=None):
    """
    Description:
    ------------
    Get the value of the slider.

    Related Pages:

			https://api.jqueryui.com/slider/#method-value

    Attributes:
    ----------
    :param jsData:
    """
    if jsData is None:
      return JsObjects.JsObjects.get('%s.slider("value")' % self._src.dom.jquery.varId)

    jsData = JsUtils.jsConvertData(jsData, None)
    return JsObjects.JsObjects.get('%s.slider("value", %s)' % (self._src.dom.jquery.varId, jsData))

  def values(self, index=0, jsData=None):
    """
    Description:
    ------------
    Get the value for the specified handle.

    Related Pages:

			https://api.jqueryui.com/slider/#method-value
    """
    if jsData is None:
      return JsObjects.JsObjects.get('%s.slider("values", %s)' % (self._src.dom.jquery.varId, index))

    jsData = JsUtils.jsConvertData(jsData, None)
    return JsObjects.JsObjects.get('%s.slider("value", %s, %s)' % (self._src.dom.jquery.varId, index, jsData))


class ProgressBar(JQueryUI):

  def destroy(self):
    """
    Description:
    ------------
    Removes the progressbar functionality completely. This will return the element back to its pre-init state.

    Related Pages:

			https://api.jqueryui.com/progressbar/#method-destroy
    """
    return JsObjects.JsObjects.get('%s.progressbar("destroy")' % self._src.dom.jquery.varId)

  def disable(self):
    """
    Description:
    ------------
    Disables the progressbar.

    Related Pages:

			https://api.jqueryui.com/progressbar/#method-disable
    """
    return JsObjects.JsObjects.get('%s.progressbar("disable")' % self._src.dom.jquery.varId)

  def enable(self):
    """
    Description:
    ------------
    Enables the progressbar.

    Related Pages:

			https://api.jqueryui.com/progressbar/#method-enable
    """
    return JsObjects.JsObjects.get('%s.progressbar("enable")' % self._src.dom.jquery.varId)

  def instance(self):
    """
    Description:
    ------------
    Retrieves the progressbar's instance object. If the element does not have an associated instance, undefined is returned.

    Related Pages:

			https://api.jqueryui.com/progressbar/#method-instance
    """
    return JsObjects.JsObjects.get('%s.progressbar("instance")' % self._src.dom.jquery.varId)

  def option(self, jsData=None, jsValue=None):
    """
    Description:
    ------------
    Gets the value currently associated with the specified optionName.
    Gets an object containing key/value pairs representing the current progressbar options hash.
    Sets the value of the progressbar option associated with the specified optionName.

    Related Pages:

			https://api.jqueryui.com/progressbar/#method-instance

    Attributes:
    ----------
    :param jsData:
    :param jsValue:
    """
    if jsData is None:
      return JsObjects.JsObjects.get('%s.progressbar("option")' % self._src.dom.jquery.varId)

    jsData = JsUtils.jsConvertData(jsData, None)
    if jsValue is None:
      return JsObjects.JsObjects.get('%s.progressbar("option", %s)' % (self._src.dom.jquery.varId, jsData))

    jsValue = JsUtils.jsConvertData(jsValue, None)
    return JsObjects.JsObjects.get('%s.progressbar("option", %s, %s)' % (self._src.dom.jquery.varId, jsData, jsValue))

  def value(self, jsValue=None):
    """
    Description:
    ------------
    Retrieves the progressbar's instance object. If the element does not have an associated instance, undefined is returned.

    Related Pages:

			https://api.jqueryui.com/progressbar/#method-instance

    Attributes:
    ----------
    :param jsValue:
    """
    if jsValue is None:
      return JsObjects.JsObjects.get('%s.progressbar("value")' % self._src.dom.jquery.varId)

    jsValue = JsUtils.jsConvertData(jsValue, None)
    return JsObjects.JsObjects.get('%s.progressbar("value", %s)' % (self._src.dom.jquery.varId, jsValue))


class Menu(JQueryUI):

  def blur(self):
    """
    Description:
    ------------
    Removes focus from a menu, resets any active element styles and triggers the menu's blur event.

    Related Pages:

			https://api.jqueryui.com/menu/#method-blur
    """
    return JsObjects.JsObjects.get('%s.menu("blur")' % self._src.dom.jquery.varId)

  def collapse(self):
    """
    Description:
    ------------
    Closes the currently active sub-menu.

    Related Pages:

			https://api.jqueryui.com/menu/#method-collapse
    """
    return JsObjects.JsObjects.get('%s.menu("collapse")' % self._src.dom.jquery.varId)

  def collapseAll(self):
    """
    Description:
    ------------
    Closes all open sub-menus.

    Related Pages:

			https://api.jqueryui.com/menu/#method-collapseAll
    """
    return JsObjects.JsObjects.get('%s.menu("collapseAll", null, true)' % self._src.dom.jquery.varId)

  def destroy(self):
    """
    Description:
    ------------
    Removes the menu functionality completely. This will return the element back to its pre-init state.

    Related Pages:

			https://api.jqueryui.com/menu/#method-destroy
    """
    return JsObjects.JsObjects.get('%s.menu("destroy")' % self._src.dom.jquery.varId)

  def disable(self):
    """
    Description:
    ------------
    Disables the menu.

    Related Pages:

			https://api.jqueryui.com/menu/#method-disable
    """
    return JsObjects.JsObjects.get('%s.menu("disable")' % self._src.dom.jquery.varId)

  def enable(self):
    """
    Description:
    ------------
    Enables the menu.

    Related Pages:

			https://api.jqueryui.com/menu/#method-disable
    """
    return JsObjects.JsObjects.get('%s.menu("enable")' % self._src.dom.jquery.varId)

  def expand(self):
    """
    Description:
    ------------
    Opens the sub-menu below the currently active item, if one exists.

    Related Pages:

			https://api.jqueryui.com/menu/#method-expand
    """
    return JsObjects.JsObjects.get('%s.menu("expand")' % self._src.dom.jquery.varId)

  def instance(self):
    """
    Description:
    ------------
    Retrieves the menu's instance object. If the element does not have an associated instance, undefined is returned.

    Related Pages:

			https://api.jqueryui.com/menu/#method-instance
    """
    return JsObjects.JsObjects.get('%s.menu("instance")' % self._src.dom.jquery.varId)

  def isFirstItem(self):
    """
    Description:
    ------------
    Returns a boolean value stating whether or not the currently active item is the first item in the menu.

    Related Pages:

			https://api.jqueryui.com/menu/#method-isFirstItem
    """
    return JsObjects.JsBoolean.JsBoolean('%s.menu("isFirstItem")' % self._src.dom.jquery.varId)

  def isLastItem(self):
    """
    Description:
    ------------
    Returns a boolean value stating whether or not the currently active item is the last item in the menu

    Related Pages:

			https://api.jqueryui.com/menu/#method-isLastItem
    """
    return JsObjects.JsBoolean.JsBoolean('%s.menu("isLastItem")' % self._src.dom.jquery.varId)

  def next(self):
    """
    Description:
    ------------
    Moves active state to next menu item.

    Related Pages:

			https://api.jqueryui.com/menu/#method-next
    """
    return JsObjects.JsObjects.get('%s.menu("next")' % self._src.dom.jquery.varId)

  def nextPage(self):
    """
    Description:
    ------------
    Moves active state to first menu item below the bottom of a scrollable menu or the last item if not scrollable.

    Related Pages:

			https://api.jqueryui.com/menu/#method-nextPage
    """
    return JsObjects.JsObjects.get('%s.menu("nextPage")' % self._src.dom.jquery.varId)

  def option(self, jsData=None, jsValue=None):
    """
    Description:
    ------------
    Gets an object containing key/value pairs representing the current menu options hash.
    Sets the value of the menu option associated with the specified optionName.
    Sets one or more options for the menu.

    Related Pages:

			https://api.jqueryui.com/progressbar/#method-option

    Attributes:
    ----------
    :param jsData:
    :param jsValue:
    """
    if jsData is None:
      return JsObjects.JsObjects.get('%s.menu("option")' % self._src.dom.jquery.varId)

    jsData = JsUtils.jsConvertData(jsData, None)
    if jsValue is None:
      return JsObjects.JsObjects.get('%s.menu("option", %s)' % (self._src.dom.jquery.varId, jsData))

    jsValue = JsUtils.jsConvertData(jsValue, None)
    return JsObjects.JsObjects.get('%s.menu("option", %s, %s)' % (self._src.dom.jquery.varId, jsData, jsValue))

  def previous(self):
    """
    Description:
    ------------
    Moves active state to previous menu item.

    Related Pages:

			https://api.jqueryui.com/menu/#method-previous
    """
    return JsObjects.JsObjects.get('%s.menu("previous")' % self._src.dom.jquery.varId)

  def previousPage(self):
    """
    Description:
    ------------
    Moves active state to first menu item above the top of a scrollable menu or the first item if not scrollable.

    Related Pages:

			https://api.jqueryui.com/menu/#method-previousPage
    """
    return JsObjects.JsObjects.get('%s.menu("previousPage")' % self._src.dom.jquery.varId)

  def refresh(self):
    """
    Description:
    ------------
    nitializes sub-menus and menu items that have not already been initialized.
    New menu items, including sub-menus can be added to the menu or all of the contents of the menu can be replaced and then initialized with the refresh() method.

    Related Pages:

			https://api.jqueryui.com/menu/#method-refresh
    """
    return JsObjects.JsObjects.get('%s.menu("refresh")' % self._src.dom.jquery.varId)

  def select(self):
    """
    Description:
    ------------
    Selects the currently active menu item, collapses all sub-menus and triggers the menu's select event.

    Related Pages:

			https://api.jqueryui.com/menu/#method-select
    """
    return JsObjects.JsObjects.get('%s.menu("select")' % self._src.dom.jquery.varId)


class Dialog(JQueryUI):

  def close(self):
    """
    Description:
    ------------
    Closes the dialog.

    Related Pages:

			https://api.jqueryui.com/dialog/#method-close
    """
    return JsObjects.JsObjects.get('%s.dialog("close")' % self._src.dom.jquery.varId)

  def destroy(self):
    """
    Description:
    ------------
    Removes the dialog functionality completely. This will return the element back to its pre-init state.

    Related Pages:

			https://api.jqueryui.com/dialog/#method-destroy
    """
    return JsObjects.JsObjects.get('%s.dialog("destroy")' % self._src.dom.jquery.varId)

  def instance(self):
    """
    Description:
    ------------
    Retrieves the dialog's instance object. If the element does not have an associated instance, undefined is returned.

    Related Pages:

			https://api.jqueryui.com/dialog/#method-instance
    """
    return JsObjects.JsObjects.get('%s.dialog("instance")' % self._src.dom.jquery.varId)

  def isOpen(self):
    """
    Description:
    ------------
    Whether the dialog is currently open.

    Related Pages:

			https://api.jqueryui.com/dialog/#method-isOpen
    """
    return JsObjects.JsObjects.get('%s.dialog("isOpen")' % self._src.dom.jquery.varId)

  def moveToTop(self):
    """
    Description:
    ------------
    Moves the dialog to the top of the dialog stack.

    Related Pages:

			https://api.jqueryui.com/dialog/#method-moveToTop
    """
    return JsObjects.JsObjects.get('%s.dialog("moveToTop")' % self._src.dom.jquery.varId)

  def open(self):
    """
    Description:
    ------------
    Opens the dialog.

    Related Pages:

			https://api.jqueryui.com/dialog/#method-open
    """
    return JsObjects.JsObjects.get('%s.dialog("open")' % self._src.dom.jquery.varId)

  def option(self, jsData=None, jsValue=None):
    """
    Description:
    ------------
    Gets the value currently associated with the specified optionName.
    Gets an object containing key/value pairs representing the current dialog options hash.
    Sets the value of the dialog option associated with the specified optionName.
    Sets one or more options for the dialog.

    Related Pages:

			https://api.jqueryui.com/dialog/#method-option

    Attributes:
    ----------
    :param jsData:
    :param jsValue:
    """
    if jsData is None:
      return JsObjects.JsObjects.get('%s.dialog("option")' % self._src.dom.jquery.varId)

    jsData = JsUtils.jsConvertData(jsData, None)
    if jsValue is None:
      return JsObjects.JsObjects.get('%s.dialog("option", %s)' % (self._src.dom.jquery.varId, jsData))

    jsValue = JsUtils.jsConvertData(jsValue, None)
    return JsObjects.JsObjects.get('%s.dialog("option", %s, %s)' % (self._src.dom.jquery.varId, jsData, jsValue))


class Autocomplete(JQueryUI):

  def close(self):
    """
    Description:
    ------------
    Closes the dialog.

    Related Pages:

			https://api.jqueryui.com/autocomplete/#method-close
    """
    return JsObjects.JsObjects.get('%s.autocomplete("close")' % self._src.dom.jquery.varId)

  def destroy(self):
    """
    Description:
    ------------
    Removes the autocomplete functionality completely. This will return the element back to its pre-init state.

    Related Pages:

			https://api.jqueryui.com/autocomplete/#method-destroy
    """
    return JsObjects.JsObjects.get('%s.autocomplete("destroy")' % self._src.dom.jquery.varId)

  def disable(self):
    """
    Description:
    ------------
    Disables the autocomplete.

    Related Pages:

			https://api.jqueryui.com/autocomplete/#method-disable
    """
    return JsObjects.JsObjects.get('%s.autocomplete("disable")' % self._src.dom.jquery.varId)

  def enable(self):
    """
    Description:
    ------------
    Enables the autocomplete.

    Related Pages:

			https://api.jqueryui.com/autocomplete/#method-enable
    """
    return JsObjects.JsObjects.get('%s.autocomplete("enable")' % self._src.dom.jquery.varId)

  def instance(self):
    """
    Description:
    ------------
    Retrieves the autocomplete's instance object. If the element does not have an associated instance, undefined is returned.

    Related Pages:

			https://api.jqueryui.com/autocomplete/#method-enable
    """
    return JsObjects.JsObjects.get('%s.autocomplete("instance")' % self._src.dom.jquery.varId)

  def option(self, jsData=None, jsValue=None):
    """
    Description:
    ------------
    Gets the value currently associated with the specified optionName.
    Gets an object containing key/value pairs representing the current dialog options hash.
    Sets the value of the dialog option associated with the specified optionName.
    Sets one or more options for the dialog.

    Related Pages:

			https://api.jqueryui.com/autocomplete/#method-option

    Attributes:
    ----------
    :param jsData:
    :param jsValue:
    """
    if jsData is None:
      return JsObjects.JsObjects.get('%s.autocomplete("option")' % self._src.dom.jquery.varId)

    jsData = JsUtils.jsConvertData(jsData, None)
    if jsValue is None:
      return JsObjects.JsObjects.get('%s.autocomplete("option", %s)' % (self._src.dom.jquery.varId, jsData))

    jsValue = JsUtils.jsConvertData(jsValue, None)
    return JsObjects.JsObjects.get('%s.autocomplete("option", %s, %s)' % (self._src.dom.jquery.varId, jsData, jsValue))

  def search(self, jsData=None):
    """
    Description:
    ------------
    Triggers a search event and invokes the data source if the event is not canceled.
    Can be used by a selectbox-like button to open the suggestions when clicked. When invoked with no parameters, the current input's value is used. Can be called with an empty string and minLength: 0 to display all items.

    Related Pages:

			https://api.jqueryui.com/autocomplete/#method-close

    Attributes:
    ----------
    :param jsData:
    """
    jsData = JsUtils.jsConvertData(jsData, None)
    return JsObjects.JsObjects.get('%s.autocomplete("search", %s)' % (self._src.dom.jquery.varId, jsData))


class Datepicker(JQueryUI):
  # TODO add dialog

  def destroy(self):
    """
    Description:
    ------------
    Removes the datepicker functionality completely. This will return the element back to its pre-init state.

    Related Pages:

			https://api.jqueryui.com/datepicker/#method-destroy
    """
    return JsObjects.JsObjects.get('%s.datepicker("destroy")' % self._src.dom.jquery.varId)

  def getDate(self):
    """
    Description:
    ------------
    Returns the current date for the datepicker or null if no date has been selected.

    Related Pages:

			https://api.jqueryui.com/datepicker/#method-getDate
    """
    return JsObjects.JsObjects.get('%s.datepicker("getDate")' % self._src.dom.jquery.varId)

  def hide(self):
    """
    Description:
    ------------
    Close a previously opened date picker.

    Related Pages:

			https://api.jqueryui.com/datepicker/#method-hide
    """
    return JsObjects.JsObjects.get('%s.datepicker("hide")' % self._src.dom.jquery.varId)

  def isDisabled(self):
    """
    Description:
    ------------
    Determine whether a date picker has been disabled.

    Related Pages:

			https://api.jqueryui.com/datepicker/#method-isDisabled
    """
    return JsObjects.JsObjects.get('%s.datepicker("isDisabled")' % self._src.dom.jquery.varId)

  def disable(self):
    """
    Description:
    ------------
    Disable the datepicker component

    """
    return JsObjects.JsObjects.get('%s.datepicker("option", "disabled", true )' % self._src.dom.jquery.varId)

  def enable(self):
    """
    Description:
    ------------
    Enable the datepciker component
    """
    return JsObjects.JsObjects.get('%s.datepicker("option", "disabled", false )' % self._src.dom.jquery.varId)

  def option(self, jsData=None, jsValue=None):
    """
    Description:
    ------------
    Gets the value currently associated with the specified optionName.
    Gets an object containing key/value pairs representing the current dialog options hash.
    Sets the value of the dialog option associated with the specified optionName.
    Sets one or more options for the dialog.

    Related Pages:

			https://api.jqueryui.com/datepicker/#method-option

    Attributes:
    ----------
    :param jsData:
    :param jsValue:
    """
    if jsData is None:
      return JsObjects.JsObjects.get('%s.datepicker("option")' % self._src.dom.jquery.varId)

    jsData = JsUtils.jsConvertData(jsData, None)
    if jsValue is None:
      return JsObjects.JsObjects.get('%s.datepicker("option", %s)' % (self._src.dom.jquery.varId, jsData))

    jsValue = JsUtils.jsConvertData(jsValue, None)
    return JsObjects.JsObjects.get('%s.datepicker("option", %s, %s)' % (self._src.dom.jquery.varId, jsData, jsValue))

  def refresh(self):
    """
    Description:
    ------------
    Redraw the date picker, after having made some external modifications.

    Related Pages:

			https://api.jqueryui.com/datepicker/#method-refresh
    """
    return JsObjects.JsObjects.get('%s.datepicker("refresh")' % self._src.dom.jquery.varId)

  def setDate(self, jsData=None,):
    """
    Description:
    ------------
    Sets the date for the datepicker. The new date may be a Date object or a string in the current date format (e.g., "01/26/2009"), a number of days from today (e.g., +7) or a string of values and periods ("y" for years, "m" for months, "w" for weeks, "d" for days, e.g., "+1m +7d"), or null to clear the selected date.

    Related Pages:

			https://api.jqueryui.com/datepicker/#method-setDate

    Attributes:
    ----------
    :param jsData:
    """
    jsData = JsUtils.jsConvertData(jsData, None)
    return JsObjects.JsObjects.get('%s.datepicker("setDate", %s)' % (self._src.dom.jquery.varId, jsData))

  def show(self):
    """
    Description:
    ------------
    Open the date picker. If the datepicker is attached to an input, the input must be visible for the datepicker to be shown.

    Related Pages:

			https://api.jqueryui.com/datepicker/#method-show
    """
    return JsObjects.JsObjects.get('%s.datepicker("show")' % self._src.dom.jquery.varId)



"""

"""

from epyk.core import html


class DropData(object):
  def __init__(self, context):
    self.context = context

  def dropfile(self, placeholder='Drop your files here', tooltip=None, report_name=None, file_type="OUTPUTS", profile=None):
    """
    Description:
    ------------
    Add an HTML component to drop files. The files will be dropped by default to the OUTPUT folder of the defined environment.
    Files will also be recorded in the database in order to ensure that those data will not be shared.
    The data sharing is and should be defined only by the user from the UI.

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlFiles.DropFile`

    Usage::

      rptObj.ui.drops.dropfile()

    Related Pages:
Attributes:
    ----------
    :param placeholder:
    :param tooltip:
    :param report_name:
    :param file_type:
    :param profile:

    :rtype: html.HtmlFiles.DropFile

    :return:
    """
    return self.context.register(html.HtmlFiles.DropFile(self.context.rptObj, placeholder, tooltip, report_name,
                                                         file_type, profile))



from . import HtmlButton
from . import HtmlContainer
from . import HtmlEvent
from . import HtmlFiles
from . import HtmlLinks
from . import HtmlImage
from . import HtmlInput
from . import HtmlList
from . import HtmlMenu
from . import HtmlOthers
from . import HtmlRadio
from . import HtmlSelect
from . import HtmlText
from . import HtmlTextComp
from . import HtmlTextEditor
from . import HtmlMedia
from . import Html
from . import HtmlPopup
from . import HtmlMessaging
from . import HtmlDates
from . import templates
from . import HtmlTrees
from . import HtmlTags
from . import HtmlStepper
from . import HtmlDrawer
from . import Header

from . import tables
from . import graph
from . import Defaults as Defaults_html



"""

"""


from epyk.core.html import Html

# The list of CSS classes
# from epyk.core.css.styles import CssGrpClsInput


class Popup(Html.Html):
  __reqCss, __reqJs = ['bootstrap', 'jquery-scrollbar'], ['jquery', 'jquery-scrollbar']
  name, category, callFnc = 'Popup Container', 'Container', 'popup'
  # _grpCls = CssGrpClsInput.CssClassPopup

  def __init__(self, report, htmlObj, title, color, width, height, withBackground, draggable, margin, profile):
    super(Popup, self).__init__(report, [], width=width[0], widthUnit=width[1], height=height[0], heightUnit=height[1],
                                profile=profile)
    self.inputs, self.height, self.width = [], "%s%s" % (height[0], height[1]) if height is not None else "100%", width
    if htmlObj is not None:
      if isinstance(htmlObj, list):
        for obj in htmlObj:
          self.__add__(obj)
      else:
        self.__add__(htmlObj)
    self._title, self.withBackground, self.draggable, self.margin = title, withBackground, draggable, margin
    if color is not None:
      self.css({"color": color})
    if self.withBackground:
      self.frameWidth = "%s%s" % (width[0], width[1])
      self.css({'width': '100%', 'position': 'fixed', 'height': '100%', 'background-color': 'rgba(0,0,0,0.4)',
                'left': 0, 'top': 0, 'margin': 'auto'})
      self.css({'display': 'none', 'z-index': '10000', 'text-align': 'center', 'padding-top': '100px',
                'padding-left': '%s%%' % self.margin, 'padding-right': '%s%%' % self.margin})
      if draggable:
        self._report.jsImports.add('jqueryui')
        self._report.cssImport.add('jqueryui')
        self._report.jsOnLoadFnc.add("$('#%s > table').draggable();" % self.htmlId)
    else:
      self.frameWidth = "100%"
      self.margin = None
      self.css({'position': 'absolute', 'margin': 0, 'padding': 0, 'display': 'none', 'z-index': '10000'})
      if draggable:
        self._report.jsImports.add('jqueryui')
        self._report.cssImport.add('jqueryui')
        self._report.jsOnLoadFnc.add("%s.draggable()" % self.jqId)
    self.set_attrs(name="name", value="report_popup")
    self._report.keyboard(27, "$('div[name=report_popup]').hide()")

  @property
  def title(self):
    if self._title is None:
      raise Exception("No title defined for this popup window")

    return "$('#title_%s span').text()" % self.htmlId

  def __add__(self, htmlObj):
    """ Add items to a container """
    htmlObj.inReport = False # Has to be defined here otherwise it is set too late
    htmlObj.container = "#%s_content" % self.htmlId
    if htmlObj.category in ['Input']:
      self.inputs.append((htmlObj.jqId, htmlObj.placeholder, htmlObj.val, htmlObj.__class__.__name__) )
    self.val.append(htmlObj)
    return self

  def jsSetInputVals(self, jsData):
    jsFncs = []
    for input in self.inputs:
      jsFncs.append("%s(%s, %s['%s'])" % (input[3], input[0], jsData, input[1]))
    return jsFncs

  def jsInputData(self, typeRow='dict'):
    if typeRow == 'dict':
      vals = []
      for input in self.inputs:
        if input[1] is not None:
          vals.append("'%s': %s" % (input[1], input[2]))
      return " [{%s}]" % ",".join(vals)

    return " [%s]" % ",".join([input[2] for input in self.inputs])

  def jsTitle(self, jsData='data', jsDataKey=None, isPyData=False, jsParse=False, jsFnc=None):
    jsData = self._jsData(jsData, jsDataKey, jsParse, isPyData, jsFnc)
    return '''$('#title_%(htmlId)s span').text(%(jsData)s)''' % {'htmlId': self.htmlId, 'jsData': jsData}

  def submit(self, jsFncs, label='Submit'):
    button = self._report.button(label)
    self.__add__(button)
    button.jsFrg('click', "%s;%s" % (";".join(jsFncs) if isinstance(jsFncs, list) else jsFncs, self.jsHide() ) )

  def show(self):
    fncs = []
    for htmlObj in self.vals:
      if getattr(htmlObj, 'dataSrc') is not None and htmlObj.dataSrc['type'] == 'script':
        fncs.append(htmlObj.jsLoadFromSrc(htmlObj.dataSrc.get('jsDataKey')))
      elif getattr(htmlObj, 'dataSrc') is not None and htmlObj.dataSrc['type'] == 'url':
        fncs.append(self._report.jsPost(htmlObj.dataSrc['url']))
    fncs.append("$('#%s').show()" % self.htmlId)
    fncs.append(self.jsResize())
    return ";".join(fncs)

  def jsResize(self):
    return '''
      var resizeEvent = window.document.createEvent('UIEvents');
      resizeEvent.initUIEvent('resize', true, false, window, 0);
      window.dispatchEvent(resizeEvent)''' % {'htmlId': self.htmlId}

  def jsShow(self, x=None, y=20, absolute=False):
    if self.margin is None:
      if x is None:
        x = - self.width / 2
      if absolute:
        return ";".join(["$('#%(htmlId)s').css({'left': %(x)s, 'top': %(y)s})" % {'htmlId': self.htmlId, 'x': x, 'y': y}, super(Popup, self).jsShow(), self.jsResize()])

      return ";".join(["$('#%(htmlId)s').css({'left': event.clientX + %(x)s, 'top': event.clientY + %(y)s})" % {
        'htmlId': self.htmlId, 'x': x, 'y': y}, super(Popup, self).jsShow(), self.jsResize()])

    #TODO: Fix Problem resizing content
    return ";".join(["$('#%(htmlId)s').css({'padding': '10%% %(margin)s%%'})" % {'htmlId': self.htmlId, 'margin': self.margin},
                    super(Popup, self).jsShow(), self.jsResize(), '''
                    if ($('#%(htmlId)s_content').height() > $(window).height()-150){
                      $('#%(htmlId)s_content').css({'height': $(window).height()-150 + 'px'})};
                    ''' % {'htmlId': self.htmlId}])

  def hide(self): return "$('#%s').hide()" % self.htmlId

  def html(self):
    """ Return the HTML display of a split container"""
    self.loadStyle()
    self.jsEvents()

    trTitle, closePopup = '', ''
    if self._title is not None:
      self._report.style.cssCls('CssPopupTableTitle')
      self._report.style.cssCls('CssPopupTableTitleContent')
      trTitle = '''
        <tr>
          <th name="title" id="title_%(htmlId)s" style="height:30px">
            <span style="margin-right:10px">%(title)s</span>
            <i onclick="$('#%(htmlId)s').hide()" style="margin:0 2px 0 2px;cursor:pointer" class="fas fa-window-close"></i>
          </th>
        </tr>''' % {'htmlId': self.htmlId, 'title': self._title}
    else:
      closePopup = '''<i onclick="$('#%(htmlId)s').hide()" style="margin:0 2px 0 2px;cursor:pointer" class="fas fa-window-close"></i>''' % {'htmlId': self.htmlId}

    content = '''
      <table id="%(htmlId)s_table" style="width:%(frameWidth)s;margin:auto">
        %(title)s
        <tr>
          <td style="padding:10px">
            <div style="width:100%%;text-align:right">%(closePopup)s</div>
            <div  class='scroll_content' id="%(htmlId)s_content" style="height:%(height)s;overflow:auto;width:100%%">%(objects)s</div>
          </td>
        </tr>
      </table>''' % {'title': trTitle, 'htmlId': self.htmlId, 'objects': "\n".join([val.html() for val in self.vals]),
                     'height': self.height, "frameWidth": self.frameWidth, 'closePopup': closePopup}

    self._report.jsOnLoadFnc.add(''' 
      $('#%(htmlId)s').on('click', function(e) {if(e.target == this) {$('#%(htmlId)s').hide()}});
      $('.scroll_content').mCustomScrollbar()''' % {"htmlId": self.htmlId})
    return '''<div %s>%s</div>''' % (self.get_attrs(pyClassNames=self.pyStyle), content)

  @property
  def val(self):
    return '$("#%s").html()' % self.htmlId

  def jsUpdate(self, data=''): return '$("#%s").html(%s)' % (self.htmlId, data)

  def to_word(self, document):
    pass



"""
Module for the HTML Media (audio and video) components
"""

import os

from epyk.core.html import Html

# The list of JSS modules
from epyk.core.js import JsUtils


class Media(Html.Html):
  name, category, callFnc = 'Video', 'Media', 'video'

  def __init__(self, report, video, path, width, height, htmlCode, profile, options):
    if path is None:
      path = "/img/%s" % report.run.report_name
      # Check if the file is available in the default directory
      # Otherwise raise an exception
      file_path = os.path.join(report.run.local_path, "static", video)
      if not os.path.exists(file_path):
        raise Exception("Missing file %s in %s" % (video, os.path.join(report.run.local_path, "static")))

    super(Media, self).__init__(report, {'path': path, 'video': video}, code=htmlCode,
                                css_attrs={"width": width, 'height': height}, profile=profile)
    self._jsStyles = options
    self.add_options(name="type", value='video/%s' % video.split(".")[-1])
    self.set_attrs(name="controls", value="controls")

  @property
  def _js__builder__(self):
    return '''
      var source = document.createElement("source");
      source.setAttribute('src', data.path +"/"+ data.video);
      for(var key in options){
        if(key === 'autoplay'){htmlObj.autoplay = options.autoplay}
        else{source.setAttribute(key, options[key])}};
      htmlObj.appendChild(source)'''

  def autoplay(self, flag=True):
    """
    Set the autoplay flag

    :param flag: Boolean. Default value is true
    """
    if flag:
      self.set_attrs(name="autoplay", value=flag)
    return self

  def __str__(self):
    if 'autoplay' in self._jsStyles:
      self.set_attrs(name="autoplay", value=JsUtils.jsConvertData(self._jsStyles["autoplay"], None))
    self.set_attrs(name="src", value=os.path.join(self.val['path'], self.val['video']))
    return '<video %s></video>' % self.get_attrs(pyClassNames=self.style.get_classes())


class Audio(Html.Html):
  name, category, callFnc = 'Video', 'Media', 'audio'

  def __init__(self, report, audio, path, width, height, htmlCode, profile, options):
    if path is None:
      path = "/img/%s" % report.run.report_name
      # Check if the file is available in the default directory
      # Otherwise raise an exception
      file_path = os.path.join(report.run.local_path, "static", audio)
      if not os.path.exists(file_path):
        raise Exception("Missing file %s in %s" % (audio, os.path.join(report.run.local_path, "static")))

    super(Audio, self).__init__(report, {'path': path, 'audio': audio}, css_attrs={"width": width, 'height': height}, code=htmlCode, profile=profile)
    self._jsStyles = options
    self.add_options(name="type", value='audio/%s' % {'mp3': 'mpeg'}.get(audio.split(".")[-1].lower(), audio.split(".")[-1]))
    self.set_attrs(name="controls", value="controls")

  def autoplay(self, flag=True):
    """
    Set the autoplay flag

    :param flag: Boolean. Default value is true
    """
    if flag:
      self.set_attrs(name="autoplay", value=flag)
    return self

  @property
  def _js__builder__(self):
    return '''
      var source = document.createElement("source");
      source.setAttribute('src', data.path +"/"+ data.audio);
      for(var key in options){
        if(key === 'autoplay'){htmlObj.autoplay = options.autoplay}
        else{source.setAttribute(key, options[key])}}; 
      htmlObj.appendChild(source)'''

  def __str__(self):
    if 'autoplay' in self._jsStyles:
      self.set_attrs(name="autoplay", value=JsUtils.jsConvertData(self._jsStyles["autoplay"], None))
    self.set_attrs(name="src", value=os.path.join(self.val['path'], self.val['audio']))
    return '<audio %(attrs)s></audio>' % {'attrs': self.get_attrs(pyClassNames=self.style.get_classes())}


class Youtube(Html.Html):
  name, category, callFnc = 'Youtube Video', 'Media', 'youtube'

  def __init__(self, report, link, width, height, htmlCode, profile, options):
    super(Youtube, self).__init__(report, link, css_attrs={"width": width, 'height': height}, code=htmlCode, profile=profile)
    self._jsStyles = options
    self._jsStyles['src'] = link

  def __str__(self):
    opts = ["%s='%s'" % (k, v) for k, v in self._jsStyles.items()]
    return '''
      <div %(attrs)s><iframe %(options)s></iframe></div>
      ''' % {'attrs': self.get_attrs(pyClassNames=self.style.get_classes()), 'link': self.val, 'options': " ".join(opts)}




from epyk.core.html.options import Options


class OptionsLi(object):
  def __init__(self, src, options):
    self.src = src
    self._li_css = options.get("li_css", {})
    self.li_class = options.get("li_class", [])

  @property
  def li_css(self):
    """
    """
    return self._li_css

  @li_css.setter
  def li_css(self, css):
    self._li_css = css

  @property
  def li_class(self):
    """
    """
    return self._li_class

  @li_class.setter
  def li_class(self, cls_names):
    self._li_class = cls_names


class OptionsTagItems(Options):

  @property
  def item_css(self):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param css: Dictionary. All the CSS attributes to add the any items
    """
    dflt_css = {"padding": '0 2px', 'border': '1px solid black', 'border-radius': '5px', 'margin': '2px'}
    return self._config_get(dflt_css)

  @item_css.setter
  def item_css(self, attrs):
    self._config(attrs)




from epyk.core.css.styles import GrpCls
from epyk.core.css import Classes
from epyk.core.css.styles.attributes import AttrInput


class ClassSlider(GrpCls.ClassHtml):
  def __init__(self, htmlObj):
    super(ClassSlider, self).__init__(htmlObj)
    self._css_ui_active = None
    self.classList['main'].add(self.cls_ui_active)

  @property
  def css(self):
    """
    Description:
    -----------
    Property to the underlying CSS definition to be added to the style HTML tag of a component

    :rtype: AttrInput.AttrInput
    """
    if self._css_struct is None:
      self._css_struct = AttrInput.AttrInput(self.htmlObj)
    return self._css_struct

  @property
  def css_class(self):
    """
    Description:
    -----------
    The internal class used to put a custom Style to this object.
    Only 1 CSS class can be added to an HTML object

    :rtype: Classes.CatalogInput.CatalogInput
    """
    if self._css_class is None:
      self._css_class = Classes.CatalogInput.CatalogInput(self.htmlObj._report, self.classList['main']).basic()
    return self._css_class

  @property
  def cls_ui_active(self):
    """
    Description:
    -----------

    :rtype: Classes.CatalogInput.CatalogInput
    """
    if self._css_ui_active is None:
      self._css_ui_active = Classes.CatalogInput.CatalogInput(self.htmlObj._report, self.classList['main']).active()
    return self._css_ui_active




from epyk.core.html import graph


class Plotly(object):
  def __init__(self, context):
    self.parent = context
    self.chartFamily = "Plotly"

  def line(self, record=None, y_columns=None, x_axis=None, title=None, profile=None, options=None,
           width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

    Related Pages:

			https://plot.ly/javascript/
    https://plot.ly/javascript/plotlyjs-function-reference/#common-parameters

    :param record:
    :param title:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    if record is None:
      record = []
      y_columns = y_columns or []
    agg_data = {}
    for rec in record:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis], 0) + float(rec[y])
    data = []
    for c in y_columns:
      series = {'x': [], 'y': []}
      for x, y in agg_data.get(c, {}).items():
        series['x'].append(x)
        series['y'].append(y)
      data.append(series)

    line_chart = graph.GraphPlotly.Line(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
    line_chart.options.responsive = True
    self.parent.context.register(line_chart)
    for d in data:
      line_chart.add_trace(d)
    return line_chart

  def bar(self, record=None, y_columns=None, x_axis=None, title=None, profile=None, options=None,
          width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

    Related Pages:

			https://plot.ly/javascript/
    https://plot.ly/javascript/plotlyjs-function-reference/#common-parameters

    :param record:
    :param title:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    if record is None:
      record = []
      y_columns = y_columns or []
    agg_data = {}
    for rec in record:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis], 0) + float(rec[y])
    data = []
    for c in y_columns:
      series = {'x': [], 'y': []}
      for x, y in agg_data.get(c, {}).items():
        series['x'].append(x)
        series['y'].append(y)
      data.append(series)

    bar_chart = graph.GraphPlotly.Bar(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
    self.parent.context.register(bar_chart)
    for d in data:
      bar_chart.add_trace(d)
    return bar_chart

  def hbar(self, record, y_columns=None, x_axis=None, title=None, profile=None, options=None,
          width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

    Related Pages:

			https://plot.ly/javascript/
    https://plot.ly/javascript/plotlyjs-function-reference/#common-parameters

    :param record:
    :param title:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    agg_data = {}
    for rec in record:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis], 0) + float(rec[y])
    data = []
    for c in y_columns:
      series = {'x': [], 'y': []}
      for x, y in agg_data.get(c, {}).items():
        series['x'].append(x)
        series['y'].append(y)
      data.append(series)

    bar_chart = graph.GraphPlotly.Bar(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
    self.parent.context.register(bar_chart)
    for d in data:
      bar_chart.add_trace(d, type='bar')
      bar_chart.data.orientation = 'h'
    return bar_chart

  def scatter(self, record=None, y_columns=None, x_axis=None, texts=None, title=None, profile=None, options=None,
              width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

    Related Pages:

			https://plot.ly/javascript/plotlyjs-function-reference/#common-parameters
    https://plot.ly/javascript/

    :param record:
    :param title:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    if record is None:
      data = []
    else:
      agg_data, pount_texts = {}, {}
      for rec in record:
        for i, y in enumerate(y_columns):
          if y in rec:
            agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis], 0) + float(rec[y])
            if texts is not None:
              pount_texts.setdefault(y, {})[rec[x_axis]] = rec[texts[i]]
      data = []
      for c in y_columns:
        series = {'x': [], 'y': [], 'text': []}
        for x, y in agg_data.get(c, {}).items():
          series['x'].append(x)
          series['y'].append(y)
          if texts is not None:
            series['text'].append(pount_texts[c][x])
        data.append(series)
    sc_chart = graph.GraphPlotly.Line(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
    self.parent.context.register(sc_chart)
    for i, d in enumerate(data):
      sc_chart.add_trace(d, mode='markers+text' if texts is not None else 'markers', type="scatter")
      sc_chart.data.marker.color = self.parent.context.rptObj.theme.colors[i]
    sc_chart.layout.no_background()
    return sc_chart

  def timeseries(self, record, y_columns=None, x_axis=None, title=None, profile=None, options=None,
                 width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

        Related Pages:

			https://plot.ly/javascript/plotlyjs-function-reference/#common-parameters
        https://plot.ly/javascript/

        :param record:
        :param title:
        :param profile:
        :param width:
        :param height:
        :param htmlCode:
        """
    agg_data = {}
    for rec in record:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis], 0) + float(rec[y])
    data = []
    for c in y_columns:
      series = {'x': [], 'y': []}
      for x, y in agg_data[c].items():
        series['x'].append(x)
        series['y'].append(y)
      data.append(series)

    sc_chart = graph.GraphPlotly.Line(self.parent.context.rptObj, width, height, title, options or {}, htmlCode,
                                      profile)
    self.parent.context.register(sc_chart)
    for d in data:
      sc_chart.add_trace(d, mode='lines', type="scatter")
    return sc_chart

  def scattergl(self, record, y_columns=None, x_axis=None, title=None, profile=None, options=None,
              width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

    Related Pages:

			https://plot.ly/javascript/plotlyjs-function-reference/#common-parameters
    https://plot.ly/javascript/

    :param record:
    :param title:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    agg_data = {}
    for rec in record:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis], 0) + float(rec[y])
    data = []
    for c in y_columns:
      series = {'x': [], 'y': []}
      for x, y in agg_data[c].items():
        series['x'].append(x)
        series['y'].append(y)
      data.append(series)

    sc_chart = graph.GraphPlotly.Line(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
    self.parent.context.register(sc_chart)
    for d in data:
      sc_chart.add_trace(d, type="scattergl", mode='markers')
    return sc_chart

  def histogram(self, record, y_columns=None, x_columns=None, title=None, profile=None, options=None,
                width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

    Related Pages:

			https://plot.ly/javascript/plotlyjs-function-reference/#common-parameters
    https://plot.ly/javascript/

    :param record:
    :param title:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:

    :return:

    :rtype: graph.GraphPlotly.Chart
    """
    data = []
    histo_axis = ('y', y_columns) if y_columns is not None else ('x', x_columns)
    for rec in record:
      for y in histo_axis[1]:
        series = {histo_axis[0]: []}
        if y in rec:
          series[histo_axis[0]].append(float(rec[y]))
        data.append(series)
    histo_chart = graph.GraphPlotly.Bar(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
    self.parent.context.register(histo_chart)

    for d in data:
      histo_chart.add_trace(d, type='histogram')
    return histo_chart

  def pie(self, record=None, y_columns=None, x_axis=None, title=None, profile=None, options=None,
              width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

    Related Pages:

			https://plot.ly/javascript/plotlyjs-function-reference/#common-parameters
    https://plot.ly/javascript/

    :param record:
    :param title:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    if record is None:
      record = []
      y_columns = y_columns or []
    agg_data = {}
    for rec in record:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis], 0) + float(rec[y])
    data = []
    for c in y_columns:
      series = {'labels': [], 'values': []}
      for x, y in agg_data.get(c, {}).items():
        series['labels'].append(x)
        series['values'].append(y)
      data.append(series)

    pie_chart = graph.GraphPlotly.Pie(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
    self.parent.context.register(pie_chart)
    for d in data:
      pie_chart.add_trace(d)
    return pie_chart

  def area(self, record, y_columns=None, x_axis=None, title=None, profile=None, options=None,
              width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

    Related Pages:

			https://plot.ly/javascript/plotlyjs-function-reference/#common-parameters
    https://plot.ly/javascript/

    :param record:
    :param title:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    agg_data = {}
    for rec in record:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis], 0) + float(rec[y])
    data = []
    for c in y_columns:
      series = {'x': [], 'y': []}
      for x, y in agg_data.get(c, {}).items():
        series['x'].append(x)
        series['y'].append(y)
      data.append(series)

    line_chart = graph.GraphPlotly.Line(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
    self.parent.context.register(line_chart)
    for d in data:
      line_chart.add_trace(d)
      line_chart.data.type = "scatter"
      line_chart.data.fill = "tozeroy"
    return line_chart

  def bubble(self, record, y_columns=None, x_axis=None, title=None, profile=None, options=None,
              width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

    Related Pages:

			https://plot.ly/javascript/plotlyjs-function-reference/#common-parameters
    https://plot.ly/javascript/

    :param data:
    :param title:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    agg_data = {}
    for rec in record:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis], 0) + float(rec[y])
    data = []
    for c in y_columns:
      series = {'x': [], 'y': []}
      for x, y in agg_data.get(c, {}).items():
        series['x'].append(x)
        series['y'].append(y)
      data.append(series)

    line_chart = graph.GraphPlotly.Line(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
    self.parent.context.register(line_chart)
    for d in data:
      line_chart.add_trace(d, mode="markers")
    return line_chart

  def ribbon(self, record, y_columns=None, x_axis=None, z_axis=None, title=None, profile=None, options=None,
              width=(100, "%"), height=(330, "px"), htmlCode=None):
    """
    Create ribbons on the x axis

    https://plot.ly/javascript/ribbon-plots/

    :param record:
    :param y_columns:
    :param x_axis:
    :param z_axis:
    :param title:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    """
    agg_data, z_data = {}, {}
    for rec in record:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis], 0) + float(rec[y])
          z_data.setdefault(y, {})[rec[x_axis]] = rec.get(z_axis)
    data = []
    for c in y_columns:
      series = {'x': [], 'y': [], 'z': []}
      for x, y in agg_data.get(c, {}).items():
        series['x'].append([x, x+1])
        series['y'].append([y, y])
        series['z'].append([z_data[c][x], z_data[c][x]])
      data.append(series)
    line_chart = graph.GraphPlotly.Surface(self.parent.context.rptObj, width, height, title, options or {}, htmlCode,
                                          profile)
    self.parent.context.register(line_chart)
    for d in data:
      line_chart.add_trace(d)
      line_chart.data.showscale = False
    return line_chart

  def surface(self, record, y_columns=None, x_axis=None, z_axis=None, title=None, profile=None,
             options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

    :param record:
    :param y_columns:
    :param x_axis:
    :param z_axis:
    :param title:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    """

    z_a, x_a, agg_y = set(), set(), {}
    for rec in record:
      if z_axis in rec:
        z_a.add(rec[z_axis])
      if x_axis in rec:
        x_a.add(rec[x_axis])
      if z_axis in rec and x_axis in rec:
        agg_key = (rec[x_axis], rec[z_axis])
        for y in y_columns:
          agg_y.setdefault(agg_key, {})[y] = agg_y.get(agg_key, {}).get(y, 0) + float(rec[y])
    z_array = sorted(list(z_a))
    x_array = sorted(list(x_a))
    naps = []
    for y in y_columns:
      nap = []
      for z in z_array:
        row = [agg_y.get((x, z), {}).get(y) for x in x_array]
        nap.append(row)
      naps.append(nap)

    surf_chart = graph.GraphPlotly.Surface(self.parent.context.rptObj, width, height, title, options or {}, htmlCode,
                                           profile)
    self.parent.context.register(surf_chart)
    for d in naps:
      surf_chart.add_trace({'z': d})
      surf_chart.data.showscale = False
    return surf_chart

  def scatter3d(self, record, y_columns=None, x_columns=None, z_columns=None, title=None, profile=None,
             options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

    https://plot.ly/javascript/3d-line-plots/

    :param record:
    :param y_columns:
    :param x_columns:
    :param z_columns:
    :param title:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    """
    x_series, y_all_series, z_series = [], [], []
    for i, y in enumerate(y_columns):
      xs, ys, zs = [], [], []
      for rec in record:
        xs.append(rec.get(x_columns[i]))
        zs.append(rec.get(z_columns[i]))
        ys.append(rec.get(y))
      x_series.append(xs)
      y_all_series.append(ys)
      z_series.append(zs)

    sc_chart = graph.GraphPlotly.Scatter3D(self.parent.context.rptObj, width, height, title, options or {}, htmlCode,
                                          profile)
    self.parent.context.register(sc_chart)
    for i, y_series in enumerate(y_all_series):
      sc_chart.add_trace({'x': x_series[i], 'y': y_series, 'z': z_series[i]})
      sc_chart.data.line.color = self.parent.context.rptObj.theme.colors[i]
    return sc_chart

  def maps(self, records, title=None, profile=None, options=None, width=(100, "%"),
           height=(330, "px"), htmlCode=None):

    surf_chart = graph.GraphPlotly.Surface(self.parent.context.rptObj, width, height, title, options or {}, htmlCode,
                                           profile)
    self.parent.context.register(surf_chart)
    for d in records:
      surf_chart.add_trace({'z': d})
    return surf_chart

  def mesh3d(self, records, intensity, x, y, z, i=None, j=None, k=None, title=None, profile=None,
             options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

    https://plot.ly/javascript/3d-mesh/

    :param records:
    :param intensity:
    :param x:
    :param y:
    :param z:
    :param i:
    :param j:
    :param k:
    :param title:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    """
    data = {"intensity": [], 'x': [], 'y': [], 'z': []}
    if i is not None:
      data[i] = []
    if j is not None:
      data[j] = []
    if k is not None:
      data[k] = []
    for rec in records:
      data["intensity"].append(rec[intensity])
      data["x"].append(rec[x])
      data["y"].append(rec[y])
      data["z"].append(rec[z])
      if i is not None:
        data["i"].append(rec[i])
      if j is not None:
        data["j"].append(rec[j])
      if k is not None:
        data["k"].append(rec[k])
    mesh_chart = graph.GraphPlotly.Mesh3d(self.parent.context.rptObj, width, height, title, options or {}, htmlCode,
                                          profile)
    self.parent.context.register(mesh_chart)
    mesh_chart.add_trace(data)
    return mesh_chart

  def number(self, value, title=None, profile=None, options=None, width=(100, "%"), height=(330, "px"),
             htmlCode=None):
    ind = graph.GraphPlotly.Indicator(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
    self.parent.context.register(ind)
    ind.add_trace({'value': value}, mode="number")
    return ind

  def number_with_delta(self, value, title=None, profile=None, options=None, width=(100, "%"),
                        height=(330, "px"), htmlCode=None):
    ind = graph.GraphPlotly.Indicator(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
    self.parent.context.register(ind)
    ind.add_trace({'value': value}, mode="number+delta")
    return ind

  def gauge(self, value, title=None, profile=None, options=None, width=(100, "%"),
                        height=(330, "px"), htmlCode=None):
    """

    gauge = rptObj.ui.charts.plotly.gauge(2000)
    gauge.data.gauge.axis.range = [0, 5000]

    https://plot.ly/javascript/indicator/

    :param value:
    :param title:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    """
    gau = graph.GraphPlotly.Indicator(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
    self.parent.context.register(gau)
    gau.add_trace({'value': value}, mode="gauge+number")
    return gau

  def scatterpolar(self, records, r_columns=None, theta_axis=None, title=None, profile=None,
             options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):

    all_series = []
    for c in r_columns:
      series = {"r": [], "theta": []}
      for rec in records:
        series['r'].append(rec[c])
        series['theta'].append(rec[theta_axis])
      all_series.append(series)
    spolar_chart = graph.GraphPlotly.ScatterPolar(self.parent.context.rptObj, width, height, title, options or {},
                                                  htmlCode, profile)
    self.parent.context.register(spolar_chart)
    for d in all_series:
      spolar_chart.add_trace(d, mode="line")
      spolar_chart.data.marker.color = None
    return spolar_chart

  def box(self, records, y_columns=None, x_columns=None, title=None, profile=None,
             options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):

    axis, cols = ('y', y_columns) if y_columns is not None else ('x', x_columns)
    series = []
    for c in cols:
      series.append([rec.get(c) for rec in records])

    box_chart = graph.GraphPlotly.Box(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
    self.parent.context.register(box_chart)
    for s in series:
      box_chart.add_trace({axis: s})
    return box_chart

  def group_box(self, records, y_columns=None, x_axis=None, title=None, profile=None,
             options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):

    """
    https://plot.ly/javascript/box-plots/

    :param records:
    :param y_columns:
    :param x_axis:
    :param title:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    """
    series, x = [[] for _ in range(len(y_columns))], []
    for rec in records:
      for i, c in enumerate(y_columns):
        series[i].append(rec.get(c))
      x.append(rec.get(x_axis))
    box_chart = graph.GraphPlotly.Box(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
    self.parent.context.register(box_chart)
    for s in series:
      box_chart.add_trace({'y': s, 'x': x})
    return box_chart

  def candlestick(self, records, closes, highs, lows, opens, x_axis, title=None, profile=None,
                  options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):

    """

    data = rptObj.py.requests.csv(data_urls.PLOTLY_APPLE_PRICES)
    sc = rptObj.ui.charts.plotly.candlestick(data, closes=["AAPL.Close"], highs=["AAPL.High"], lows=["AAPL.Low"], opens=["AAPL.Open"], x_axis='Date')

    :param records:
    :param closes:
    :param highs:
    :param lows:
    :param opens:
    :param x_axis:
    :param title:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    :return:
    """
    all_series = []
    for i, c in enumerate(closes):
      series = {'x': [], 'close': [], 'high': [], 'low': [], 'open': []}
      for rec in records:
        series['x'].append(rec[x_axis])
        series['close'].append(rec[c])
        series['high'].append(rec[highs[i]])
        series['low'].append(rec[lows[i]])
        series['open'].append(rec[opens[i]])
      all_series.append(series)

    candle_chart = graph.GraphPlotly.CandleStick(self.parent.context.rptObj, width, height, title, options or {}, htmlCode,
                                       profile)
    self.parent.context.register(candle_chart)
    for s in all_series:
      candle_chart.add_trace(s)
    candle_chart.layout.no_background()
    return candle_chart
